from unittest.mock import patch

import pytest

from harp.utils.env import cast_bool, get_bool_from_env

parametrize_with_string_booleans = pytest.mark.parametrize(
    "value, expected",
    [
        ("true", True),
        ("yes", True),
        ("1", True),
        ("false", False),
        ("no", False),
        ("0", False),
    ],
)


@parametrize_with_string_booleans
def test_cast_bool(value, expected):
    assert cast_bool(value) == expected


@parametrize_with_string_booleans
def test_get_bool_from_env(value, expected):
    with patch.dict("os.environ", {"TEST_ENV_VAR": value}):
        assert get_bool_from_env("TEST_ENV_VAR", False) == expected



import pytest
from pydantic import ValidationError

from harp.config.asdict import asdict
from harp.utils.collections import all_combinations
from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.proxy.constants import AVAILABLE_POOLS
from harp_apps.proxy.settings.remote import RemoteEndpointSettings
from harp_apps.proxy.settings.remote.endpoint import RemoteEndpoint


class TestRemoteEndpointSettings(BaseConfigurableTest):
    type = RemoteEndpointSettings
    initial = {"url": "http://example.com"}
    expected = {"url": "http://example.com/"}
    expected_verbose = {
        "url": "http://example.com/",
        "pools": ["default"],
        "liveness": {"type": "inherit"},
    }

    @pytest.mark.parametrize("pools", all_combinations(AVAILABLE_POOLS))
    def test_valid_pools(self, pools):
        obj = self.create(pools=pools)
        assert asdict(obj) == {**self.expected, "pools": list(sorted(pools))}

    def test_invalid_pool(self):
        invalid_data = {"url": "http://example.com", "pools": ["invalid"]}
        with pytest.raises(ValueError):
            RemoteEndpointSettings(**invalid_data)

        with pytest.raises(ValidationError) as exc_info:
            RemoteEndpointSettings.model_validate(invalid_data)

        errors = exc_info.value.errors()
        assert errors[0]["loc"] == ("pools",)
        assert errors[0]["msg"] == "Value error, Invalid pool names: invalid."

    def test_valid_thresholds(self):
        valid_data = {
            "url": "http://example.com",
            "liveness": {
                "type": "naive",
                "failure_threshold": 3,
                "success_threshold": 5,
            },
        }
        settings = RemoteEndpointSettings(**valid_data)
        assert settings.liveness.failure_threshold == 3
        assert settings.liveness.success_threshold == 5

    def test_asdict(self):
        endpoint = self.create(url="http://example.com")
        expected = {
            "url": "http://example.com/",
        }
        assert asdict(endpoint) == expected

        # idempotence
        endpoint = self.create(**asdict(endpoint))
        assert asdict(endpoint) == expected

    def test_asdict_with_nondefault_thresholds(self):
        endpoint = self.create(
            url="http://example.com", liveness={"type": "naive", "success_threshold": 2, "failure_threshold": 4}
        )
        expected = {
            "url": "http://example.com/",
            "liveness": {
                "type": "naive",
                "failure_threshold": 4,
                "success_threshold": 2,
            },
        }
        assert asdict(endpoint) == expected

        # idempotence
        endpoint = RemoteEndpointSettings(**asdict(endpoint))
        assert asdict(endpoint) == expected


class TestRemoteEndpointStateful(BaseConfigurableTest):
    type = RemoteEndpoint
    initial = {"settings": TestRemoteEndpointSettings.initial}
    expected = {
        "settings": {
            "pools": ["default"],
            "url": "http://example.com/",
            "liveness": {"type": "inherit"},
        },
    }
    expected_verbose = {
        "settings": TestRemoteEndpointSettings.expected_verbose,
        "failure_reasons": None,
        "status": 0,
    }



import re

from colorama import Fore
from honcho.process import Process


class QuietViteHonchoProcess(Process):
    """
    This is a quick hack to avoid misleading log at start of internal vite dev serversubprocess.

    """

    _muted = True
    _status = ""

    def _send_message(self, data, type="line"):
        if type == "line" and self._muted:
            if b"  VITE v" in data:
                self._status = data.decode("utf-8").strip()

            if b"Local:   http://localhost:" in data:
                url = re.search("(?P<url>https?://[^\s]+)", data.decode("utf-8")).group("url")
                super()._send_message(
                    (
                        "📈  "
                        + Fore.LIGHTBLUE_EX
                        + "Dashboard development server started."
                        + Fore.RESET
                        + (f" ({self._status})" if self._status else "")
                    ).encode(),
                )
                super()._send_message(("  ➜ Internal url: " + url).encode())
                super()._send_message(
                    "  ➜ This url is for internal use only, you should use the proxied url instead.".encode()
                )

            if b"press h + enter to show help" in data:
                self._muted = False
            return

        return super()._send_message(data, type)



from collections import ChainMap
from itertools import chain, combinations

from multidict import CIMultiDict

from harp.typing import NotSet


class MultiChainMap(ChainMap):
    """
    Implements the logic of chainmap, but for multidicts. It also allows to unset values on some key, this will be used
    for headers, cookies, query string ... to allow the user to rewrite requests/responses in a memory efficient way.

    The current implementation is neither complete, nor optimized.

    TODO: find a better way for __iter__
    """

    def __getitem__(self, key):
        x = super().__getitem__(key)
        if x is NotSet:
            return self.__missing__(key)
        return x

    def __delitem__(self, key):
        if key not in self:
            return super().__delitem__(key)
        self.maps[0][key] = NotSet

    def __contains__(self, key):
        if key in self.maps[0] and self.maps[0][key] is NotSet:
            return False
        return super().__contains__(key)

    def __len__(self):
        return len(list(self.__iter__()))

    def items(self):
        # TODO this needs a better algorithm (that does not involve creating a lot of multidicts)
        d = CIMultiDict()
        for mapping in reversed(self.maps):
            d.update(CIMultiDict(((k, v) for k, v in mapping.items())))
        return [(k, v) for k, v in d.items() if v is not NotSet]

    def __iter__(self):
        # TODO this needs a better algorithm (that does not involve creating a lot of multidicts)
        d = CIMultiDict()
        for mapping in reversed(self.maps):
            d.update(CIMultiDict(((k, v) for k, v in mapping.items())))
        return iter(CIMultiDict(((k, v) for k, v in d.items() if v is not NotSet)))

    def pop(self, key):
        # it has already been unset, or is not there
        if not self.__contains__(key):
            raise KeyError(f"Key not found in the first mapping: {key!r}")

        # otherwise, it's simpler to just set it as not set
        value = self[key]
        self[key] = NotSet
        return value

    def popitem(self):
        raise NotImplementedError()


def all_combinations(iterable, *, maxlen=None):
    return set(
        chain(
            *(
                combinations(iterable, n + 1)
                for n in range(min(maxlen, len(iterable)) if maxlen and (maxlen > 0) else len(iterable))
            )
        )
    )



from harp_apps.dashboard.filters.utils import str_to_float_or_none


def test_str_to_float_or_none():
    assert str_to_float_or_none("123.45") == 123.45
    assert str_to_float_or_none("0") == 0.0
    assert str_to_float_or_none("-123.45") == -123.45
    assert str_to_float_or_none("not a number") is None
    assert str_to_float_or_none("") is None



from math import log, sqrt


def _tpdex(n):
    # This formula looks like a mess, but we spent quite some time for tuning it. Please do not change this unless
    # you are aware of the previous work and the impact it has.
    if n < 1328:
        return min(1 - 7.22e-4 * n + 2.44e-7 * (n**2) - 3e-11 * (n**3), 1)
    return max(-0.42115982438869104 * log(sqrt(n) + 1) + 1.9270515661020815, 0)


def tpdex(n):
    return int(_tpdex(n) * 100)



from copy import deepcopy

import pytest

from harp.config import ConfigurationBuilder
from harp.controllers import ProxyControllerResolver

from ..controllers.system import SystemController


class SystemControllerTestFixtureMixin:
    @pytest.fixture
    async def controller(self, request, sql_storage, blob_storage):
        # retrieve settings overrides
        try:
            raw_settings = deepcopy(request.param)
        except AttributeError:
            raw_settings = {}
        raw_settings.setdefault("applications", [])
        if "storage" not in raw_settings["applications"]:
            raw_settings["applications"].append("storage")

        raw_settings.setdefault("storage", {})

        db_url = sql_storage.engine.url.render_as_string(hide_password=False)
        if raw_settings["storage"].get("url") and raw_settings["storage"]["url"] != db_url:
            raise ValueError(f"Database URL mismatch: {raw_settings['storage']['url']} != {db_url}")
        raw_settings["storage"]["url"] = db_url

        raw_settings["storage"].setdefault("blobs", {})
        raw_settings["storage"]["blobs"].setdefault("type", blob_storage.type)

        system = await ConfigurationBuilder(raw_settings, use_default_applications=False).abuild_system()

        try:
            yield SystemController(
                storage=sql_storage,
                settings=system.config,
                handle_errors=False,
                resolver=system.provider.get(ProxyControllerResolver),
            )
        finally:
            await system.dispose()


def parametrize_with_settings(*args):
    return pytest.mark.parametrize("controller", args, indirect=True)



def test_exposed_api(snapshot):
    """This test snapshots the public API of the harp.config module, to make sure that api changes are conscious
    choices and not coincidences."""
    api = __import__("harp.config", fromlist=["*"])

    api_elements = []
    for element in dir(api):
        if element not in api.__all__:
            continue

        api_elements.append(element)

    assert api_elements == snapshot



"""
The Config (:mod:`harp.config`) module provides tools to configure *Applications*.

*Applications* are standard python packages that contains specific definitions to be able to register services and
configure themselves. Every feature of Harp is built as an application, and you will want to write applications if you
want to extend harp's functionality for specific needs.


Contents
--------

"""

__title__ = "Config"

from .applications import Application, ApplicationsRegistry
from .asdict import asdict
from .builders import ConfigurationBuilder
from .configurables import Configurable, LazyService, Service, Stateful
from .events import (
    EVENT_BIND,
    EVENT_BOUND,
    EVENT_READY,
    EVENT_SHUTDOWN,
    OnBindEvent,
    OnBoundEvent,
    OnReadyEvent,
    OnShutdownEvent,
)

__all__ = [
    "Application",
    "ApplicationsRegistry",
    "Configurable",
    "ConfigurationBuilder",
    "EVENT_BIND",
    "EVENT_BOUND",
    "EVENT_READY",
    "EVENT_SHUTDOWN",
    "LazyService",
    "OnBindEvent",
    "OnBoundEvent",
    "OnReadyEvent",
    "OnShutdownEvent",
    "Service",
    "Stateful",
    "asdict",
]



import os
from typing import Iterable, Self, cast

import orjson
from config.common import ConfigurationBuilder as BaseConfigurationBuilder
from config.common import MapSource, merge_values
from config.env import EnvVars

from harp.typing import GlobalSettings
from harp.utils.config.yaml import include_constructor  # noqa

from ..applications import ApplicationsRegistry
from ..defaults import DEFAULT_APPLICATIONS, DEFAULT_SYSTEM_CONFIG_FILENAMES
from ..examples import get_example_filename
from .system import System


def _get_system_configuration_sources():
    for _candidate in DEFAULT_SYSTEM_CONFIG_FILENAMES:
        if os.path.exists(_candidate):
            from config.yaml import YAMLFile

            yield YAMLFile(_candidate)
            break


class ConfigurationBuilder(BaseConfigurationBuilder):
    """
    A builder class for assembling the global configuration settings for HARP from various sources.

    This class extends config.ConfigurationBuilder, incorporating additional functionality specific to HARP,
    such as handling default applications and integrating with the ApplicationsRegistry. It supports adding
    configuration from files, environment variables, and direct values, with a focus on flexibility and ease of use.

    Attributes:
        _defaults (dict): Default values for the configuration, typically loaded from internal defaults or specified by the user.
        applications (ApplicationsRegistryType): An instance of ApplicationsRegistry or a subclass, managing the registration and configuration of HARP applications.
        applications_registry_type (type): The type of ApplicationsRegistry to use for managing applications.

    Methods:
        add_file(filename: str): Adds a single configuration file by its filename.
        add_files(filenames: Iterable[str]): Adds multiple configuration files by their filenames.
        add_values(values: dict): Adds configuration values directly from a dictionary.
        normalize(x: dict): Normalizes the configuration values, potentially transforming them based on application-specific logic.
        build() -> GlobalSettings: Constructs the final, aggregated configuration settings as a GlobalSettings instance.
        from_commandline_options(options): Class method to create an instance of ConfigurationBuilder from command line options.
        from_bytes(serialized: bytes, **kwargs) -> Self: Class method to create an instance of ConfigurationBuilder from serialized bytes.

    The ConfigurationBuilder is central to the dynamic configuration system in HARP, allowing configurations to be built
    and modified in a flexible and intuitive manner.

    """

    def __init__(
        self,
        default_values=None,
        /,
        *,
        use_default_applications=True,
    ) -> None:
        """
        Initializes a new instance of the ConfigurationBuilder.

        Parameters:
            default_values (dict, optional): A dictionary of default configuration values. Defaults to None.
            use_default_applications (bool, optional): Whether to automatically include default HARP applications in the configuration. Defaults to True.
        """
        self._defaults = default_values or {}
        self.applications = self.create_application_registry()
        self.applications_registry_type = type(self.applications)

        for app_name in self._defaults.pop("applications", []):
            self.applications.add(app_name)

        if use_default_applications:
            self.applications.add(*DEFAULT_APPLICATIONS)

        super().__init__()

    def create_application_registry(self):
        return ApplicationsRegistry()

    def add_file(self, filename: str):
        """
        Adds a configuration file to the builder.

        Parameters:
            filename (str): The path to the configuration file to add.

        Raises:
            ValueError: If the file extension is not recognized.
        """
        _, ext = os.path.splitext(filename)
        if ext in (".yaml", ".yml"):
            from config.yaml import YAMLFile

            self.add_source(YAMLFile(filename))
        elif ext in (".json",):
            from config.json import JSONFile

            self.add_source(JSONFile(filename))
        elif ext in (".ini", ".conf"):
            from config.ini import INIFile

            self.add_source(INIFile(filename))
        elif ext in (".toml",):
            from config.toml import TOMLFile

            self.add_source(TOMLFile(filename))
        else:
            raise ValueError(f"Unknown file extension: {ext}")

    def add_files(self, filenames: Iterable[str]):
        """
        Adds multiple configuration files to the builder.

        Parameters:
            filenames (Iterable[str]): An iterable of file paths to add.
        """
        for filename in filenames or ():
            self.add_file(filename)

    def add_values(self, values: dict):
        """
        Adds configuration values directly from a dictionary.

        Parameters:
            values (dict): A dictionary of configuration values to add.
        """
        # TODO: split first key on dots, with quote escaping, and create a recursive dict to apply correct merging.
        for k, v in values.items():
            self.add_value(k, v)

    def normalize(self, x: dict):
        """
        Normalizes the configuration values, potentially transforming them based on application-specific logic.

        Parameters:
            x (dict): The configuration values to normalize.

        Returns:
            dict: The normalized configuration values.
        """
        # todo: support recursive doted notation key. The easiest way would probably be to convert "a.b": ... into
        #  a: {b: ...}, meanwhile, let's be carfeul with those keys.
        return {k: (self.applications[k].normalize(v) if k in self.applications else v) for k, v in x.items()}

    def build(self) -> GlobalSettings:
        """
        Constructs the final, aggregated configuration settings as a GlobalSettings instance.

        Returns:
            GlobalSettings: The aggregated global settings derived from all added sources.
        """
        settings = {}
        for source in (
            EnvVars(prefix="DEFAULT__HARP_"),
            MapSource(self.applications.defaults()),
            MapSource(self._defaults or {}),
            *_get_system_configuration_sources(),
            *self._sources,
        ):
            merge_values(settings, self.normalize(source.get_values()))

        all_settings = []
        for name, application in self.applications.items():
            settings_type = self.applications[name].settings_type
            if not settings_type:
                continue
            _local_settings = settings.get(name.rsplit(".", 1)[-1], {})
            if not isinstance(_local_settings, settings_type):
                _local_settings = settings_type(**_local_settings)
            all_settings.append((name, _local_settings))

        return cast(
            GlobalSettings,
            {
                "applications": self.applications.aslist(),
                **{name: value for name, value in sorted(all_settings) if value},
            },
        )

    def __call__(self) -> GlobalSettings:
        return self.build()

    async def abuild_system(self) -> System:
        from .system import SystemBuilder

        return await SystemBuilder(self.applications, self.build).abuild()

    @classmethod
    def from_commandline_options(cls, options) -> Self:
        """
        Creates an instance of ConfigurationBuilder from command line options.

        Parameters:
            options: The command line options to use for building the configuration.

        Returns:
            ConfigurationBuilder: An instance of ConfigurationBuilder configured according to the provided command line options.
        """
        # todo: config instead of sources in constructor ? for example no_default_apps, etc.

        try:
            applications = options.applications
        except AttributeError:
            applications = None
        builder = cls(
            {"applications": applications} if applications else None,
            use_default_applications=not applications,
        )

        # todo: raise if enabling AND disabling an app at the same time? maybe not but instructions should be taken in
        #  order, which looks hard to do using click...
        for _enabled_application in options.enable or ():
            builder.applications.add(_enabled_application)
        for _disabled_application in options.disable or ():
            builder.applications.remove(_disabled_application)

        builder.add_files((get_example_filename(example) for example in options.examples))
        builder.add_files(options.files or ())
        builder.add_source(EnvVars(prefix="HARP_"))
        builder.add_values(options.options or {})

        _endpoints = []
        for k, v in (options.endpoints or {}).items():
            _port, _url = v.split(":", 1)
            _endpoints.append({"name": k, "port": int(_port), "url": _url})
        if len(_endpoints):
            builder.add_value("proxy.endpoints", _endpoints)

        return builder

    @classmethod
    def from_bytes(cls, serialized: bytes, **kwargs) -> Self:
        """
        Creates an instance of ConfigurationBuilder from serialized bytes.

        Parameters:
            serialized (bytes): The serialized configuration data.
            **kwargs: Additional keyword arguments to pass to the constructor.

        Returns:
            ConfigurationBuilder: An instance of ConfigurationBuilder initialized with the deserialized configuration data.
        """
        unserialized = orjson.loads(serialized)
        return cls(unserialized, use_default_applications=False, **kwargs)



import binascii
from base64 import b64decode
from typing import Optional

from httpx import AsyncByteStream, ByteStream
from multidict import CIMultiDict

from .typing import BaseHttpMessage, HttpRequestBridge
from .utils.cookies import parse_cookie


class HttpRequest(BaseHttpMessage):
    kind = "request"

    def __init__(
        self,
        impl: Optional[HttpRequestBridge] = None,
        *,
        extensions: Optional[dict] = None,
        **kwargs,
    ):
        super().__init__(extensions=extensions)

        if impl is None:
            from .tests.stubs import HttpRequestStubBridge

            impl = HttpRequestStubBridge(**kwargs)
        else:
            if len(kwargs) > 0:
                raise ValueError("Cannot pass both an implementation and non-explicit keyword arguments.")

        self._impl = impl

        # Initialize properties from the implementation bridge
        self.method = self._impl.get_method()
        self.path = self._impl.get_path()
        self.query = self._impl.get_query()
        self.server_ipaddr = self._impl.get_server_ipaddr()
        self.server_port = self._impl.get_server_port()
        self._headers = self._impl.get_headers()
        self._stream = self._impl.get_stream()

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, stream: AsyncByteStream):
        self._stream = stream
        if hasattr(self, "_body"):
            delattr(self, "_body")

    @property
    def headers(self) -> CIMultiDict:
        return self._headers

    @headers.setter
    def headers(self, headers):
        self._headers = CIMultiDict(headers)

    @property
    def cookies(self) -> dict:
        cookies = {}
        for header in self.headers.getall("cookie", []):
            cookies.update(parse_cookie(header))
        return cookies

    @property
    def basic_auth(self) -> tuple[str, str] | None:
        """Parse and returns basic auth from headers."""
        for header in self.headers.getall("authorization", []):
            try:
                _type, _auth = header.split(" ", 1)
            except ValueError:
                continue

            if _type.lower() == "basic":
                try:
                    user, passwd = b64decode(_auth).decode("utf-8").split(":", 1)
                except (binascii.Error, ValueError):
                    continue

                return user, passwd

    @property
    def body(self) -> bytes:
        """Returns the previously read body of the request. Raises a RuntimeError if the body has not been read yet,
        you must await the `read()` asynchronous method first, which cannot be done here because properties are
        synchronous, so we let the user explicitely call it before."""
        if not hasattr(self, "_body"):
            raise RuntimeError("Request body has not been read yet, please await `read()` first.")
        return self._body

    async def aread(self) -> bytes:
        """Read all chunks from request. We may want to be able to read partial body later, but for now it's all or
        nothing. This method does nothing if the body has already been read."""
        if not hasattr(self, "_body"):
            self._body = b"".join([part async for part in self._stream])
            # self.headers["content-length"] = str(len(self.body))
        if not isinstance(self._stream, ByteStream):
            self._stream = ByteStream(self._body)
        return self.body

    def __str__(self):
        return f"{self.method} {self.path}"



from typing import override

from redis.asyncio import Redis

from harp.models import Blob
from harp_apps.storage.types import IBlobStorage


class RedisBlobStorage(IBlobStorage):
    type = "redis"

    def __init__(self, client: Redis):
        self.client = client

    @override
    async def get(self, blob_id):
        data = await self.client.get(blob_id)
        if data is not None:
            content_type, body = data.split(b"\n", 1)
            return Blob(
                id=blob_id,
                data=body,
                content_type=content_type.decode() if content_type else None,
            )

    @override
    async def put(self, blob: Blob) -> Blob:
        await self.client.set(blob.id, ((blob.content_type or "") + "\n").encode() + blob.data)
        return blob

    force_put = put

    @override
    async def delete(self, blob_id):
        await self.client.delete(blob_id)

    @override
    async def exists(self, blob_id: str) -> bool:
        return await self.client.exists(blob_id)



from datetime import date, datetime

from harp.utils.dates import ensure_date, ensure_datetime


def test_ensure_date():
    assert ensure_date("2020-01-01") == date(2020, 1, 1)
    assert ensure_date(date(2020, 1, 1)) == date(2020, 1, 1)
    assert ensure_date(datetime(2020, 1, 1, 13, 37, 42)) == date(2020, 1, 1)
    assert ensure_date(None) is None


def test_ensure_datetime():
    assert ensure_datetime("2020-01-01 13:37:42.123456") == datetime(2020, 1, 1, 13, 37, 42, 123456)
    assert ensure_datetime(date(2020, 1, 1)) == datetime(2020, 1, 1)
    assert ensure_datetime(datetime(2020, 1, 1, 13, 37, 42)) == datetime(2020, 1, 1, 13, 37, 42)
    assert ensure_datetime(None) is None



import builtins
import importlib.util

code = None

if getattr(builtins, "__sphinx__", False):

    def code(x):
        return ":code:`" + x + "`"


try:
    import rich_click as click

    click.rich_click.USE_RICH_MARKUP = True

    if not code:

        def code(x):
            return "[code]" + x + "[/]"

except ImportError:
    import click

if not code:

    def code(x):
        return x


def assert_package_is_available(package_name: str):
    if importlib.util.find_spec(package_name) is None:
        raise ModuleNotFoundError(f"Package {package_name!r} is not available.")


def check_packages(*pkgs):
    for pkg in pkgs:
        try:
            assert_package_is_available(pkg)
        except ModuleNotFoundError:
            return False
    return True


__all__ = [
    "assert_package_is_available",
    "check_packages",
    "click",
]



import functools
import itertools

import pytest

from harp.utils.env import get_bool_from_env

# Setting to chose whether to test an extensive set of underlying database systems or a representing subset. Generally,
# unless you are testing a complex feature or running a final barier continuous nintegration system, the subset should
# be enough (and faster).
TEST_ALL_DATABASES = get_bool_from_env("TEST_ALL_DATABASES", False)

# Supported databases, containers and their drivers. This will be used as a matrix to parametrize tests.
TEST_DATABASES_SOURCES = {
    "sqlite": {"drivers": ["aiosqlite"]},
    "postgresql": {
        "images": [
            *(
                [
                    "postgres:12",
                    "postgres:12-alpine",
                    "postgres:13",
                    "postgres:13-alpine",
                    "postgres:14",
                    "postgres:14-alpine",
                    "postgres:15",
                    "postgres:15-alpine",
                    "postgres:16",
                    "postgres:16-alpine",
                    "timescale/timescaledb:latest-pg13",
                    "timescale/timescaledb:latest-pg14",
                    "timescale/timescaledb:latest-pg15",
                    "timescale/timescaledb:latest-pg16",
                ]
                if TEST_ALL_DATABASES
                else [
                    "postgres:16-alpine",
                ]
            ),
        ],
        "drivers": ["asyncpg"],
    },
    "mysql": {
        "images": (
            [
                "mysql:8-oracle",
                "mariadb:lts",
                "mariadb:latest",
            ]
            if TEST_ALL_DATABASES
            else ["mariadb:lts"]
        ),
        "drivers": ["aiomysql", "asyncmy"] if TEST_ALL_DATABASES else [],
    },
    # disabled, does not install cleanly on osx+arm
    # see https://github.com/pymssql/pymssql/issues/769
    # supporting microsoft shit is not a priority unless someone wants to sponsor it
    # "mssql": {"drivers": ["aioodbc"]},
}

# Combine supported databases into something our fixtures can use.
TEST_DATABASES = list(
    map(
        "|".join,
        itertools.chain(
            *(
                itertools.product(
                    [dialect],
                    TEST_DATABASES_SOURCES[dialect].get("images", [""]),
                    TEST_DATABASES_SOURCES[dialect]["drivers"],
                )
                for dialect in TEST_DATABASES_SOURCES.keys()
            )
        ),
    )
)


def parametrize_with_database_urls(*databases):
    """
    Decorator to parametrize with all supported databases (no parameters) or a subset of supported databases.

    Usage::

    @parametrize_with_database_urls():
    def test_all_databases(database_url):
        ... # will receive all supported databases urls, while a matching container runs

    @parametrize_with_database_urls("postgresql", "mysql")
    def test_all_databases(database_url):
        ... # will receive supported databases urls matching parameters, while a matching container runs
    """

    def wrapper(func):
        @pytest.mark.parametrize(
            "database_url",
            (
                [x for x in TEST_DATABASES if any(x.startswith(db + "|") for db in databases)]
                if len(databases)
                else TEST_DATABASES
            ),
            indirect=True,
        )
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            return await func(*args, **kwargs)

        return inner

    return wrapper


BLOB_STORAGE_TYPES = ["sql", "redis"]


def parametrize_with_blob_storages_urls(*types):
    """
    Decorator to parametrize with all supported blob storages.

    Usage::

    @parametrize_with_blob_storages()
    def test_all_blob_storages(blob_storage):
        ... # will receive all supported blob storages, while a matching container runs
    """

    if len(types) and not all(x in BLOB_STORAGE_TYPES for x in types):
        raise ValueError(f"Unsupported blob storage types: {types}")

    def wrapper(func):
        @pytest.mark.parametrize("blob_storage", types if len(types) else BLOB_STORAGE_TYPES, indirect=True)
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            return await func(*args, **kwargs)

        return inner

    return wrapper



from unittest.mock import ANY, Mock

import orjson
import pytest
from freezegun import freeze_time
from multidict import MultiDict

from harp.http import HttpRequest
from harp.utils.testing.communicators import ASGICommunicator
from harp.utils.testing.mixins import ControllerThroughASGIFixtureMixin
from harp_apps.dashboard.controllers.transactions import TransactionsController
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class TransactionsControllerTestFixtureMixin:
    @pytest.fixture
    def controller(self, sql_storage):
        return TransactionsController(storage=sql_storage, handle_errors=False)


class TestTransactionsController(TransactionsControllerTestFixtureMixin, StorageTestFixtureMixin):
    async def test_filters_using_handler(self, controller: TransactionsController):
        request = Mock(spec=HttpRequest, query=MultiDict())
        response = await controller.filters(request)

        # todo this format may/will change, but we add this test to ensure we start to be meticulous about quality
        assert response == {
            "endpoint": {"current": None, "values": ANY},
            "flag": {"current": None, "values": ANY, "fallbackName": ANY},
            "method": {"current": None, "values": ANY},
            "status": {"current": None, "values": ANY},
            "tpdex": {"current": {"min": ANY, "max": ANY}, "values": ANY},
        }

    async def test_filters_meta_updated(self, controller: TransactionsController):
        with freeze_time("2024-01-01 12:00:00"):
            await self.create_transaction(controller.storage, endpoint="foo")

            request = Mock(spec=HttpRequest, query=MultiDict())
            response = await controller.filters(request)

            assert response["endpoint"]["values"] == [{"count": 1, "name": "foo"}]

            # Create two more transactions
            await self.create_transaction(controller.storage, endpoint="foo")
            await self.create_transaction(controller.storage, endpoint="foo")

        # Check the count again
        with freeze_time("2024-01-01 12:00:20"):
            # Check the count again
            response = await controller.filters(request)
            assert response["endpoint"]["values"] == [{"count": 1, "name": "foo"}]

        # Move forward in time by 2 minutes
        with freeze_time("2024-01-01 12:02:00"):
            # Check the count again
            response = await controller.filters(request)
            assert response["endpoint"]["values"] == [{"count": 3, "name": "foo"}]


class TestTransactionsControllerThroughASGI(
    TransactionsControllerTestFixtureMixin,
    StorageTestFixtureMixin,
    ControllerThroughASGIFixtureMixin,
):
    async def test_filters_using_asgi(self, client: ASGICommunicator):
        # todo this is probably not the right place, we're unit testing a controller, should we really duplicate tests
        # to also go through asgi ? Let's think about this, there is no real value of aving two tests doing the same
        # with an additional layer here.
        response = await client.http_get("/api/transactions/filters")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)

        # todo this format may/will change, but we add this test to ensure we start to be meticulous about quality
        assert orjson.loads(response["body"]) == {
            "endpoint": {"current": None, "values": ANY},
            "flag": {"current": None, "values": ANY, "fallbackName": ANY},
            "method": {"current": None, "values": ANY},
            "status": {"current": None, "values": ANY},
            "tpdex": {"current": {"min": ANY, "max": ANY}, "values": ANY},
        }



from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class TestStorageMessages(StorageTestFixtureMixin):
    async def test_create_message(self, sql_storage):
        transaction = await self.create_transaction(sql_storage)
        message = await self.create_message(
            sql_storage, transaction_id=transaction.id, kind="misc", summary="bal", headers="foo", body="foo"
        )
        assert message.id is not None
        assert message.transaction_id == transaction.id
        assert message.kind == "misc"
        assert message.summary == "bal"
        assert message.headers == "foo"
        assert message.body == "foo"

    async def test_create_message_with_long_summary(self, sql_storage):
        transaction = await self.create_transaction(sql_storage)
        message = await self.create_message(
            sql_storage, transaction_id=transaction.id, kind="misc", summary="a" * 500, headers="foo", body="foo"
        )
        assert message.id is not None
        assert message.transaction_id == transaction.id
        assert message.kind == "misc"
        assert message.summary == "a" * 500
        assert message.headers == "foo"
        assert message.body == "foo"



from typing import Protocol

from httpx import AsyncByteStream
from multidict import CIMultiDict, MultiDict


class HttpRequestBridge(Protocol):
    """
    The HttpRequestBridge protocol defines the methods required by the HttpRequest object for it to attach to a real
    implementation, such as WSGI, ASGI, ...
    """

    def get_server_ipaddr(self) -> str: ...

    def get_server_port(self) -> int: ...

    def get_method(self) -> str: ...

    def get_path(self) -> str: ...

    def get_query(self) -> MultiDict: ...

    def get_headers(self) -> CIMultiDict: ...

    async def aread(self) -> bytes: ...

    def get_stream(self) -> AsyncByteStream: ...


class HttpResponseBridge(Protocol):
    """
    The HttpResponseBridge protocol defines the necessary methods to actually send an HttpResponse through a real
    channel.
    """



from typing import Literal, Tuple, get_args

TAlgorithm = Literal[
    "argon2",
    "bcrypt",
    "bcrypt_sha256",
    "cisco_asa",
    "cisco_pix",
    "cisco_type7",
    "django_bcrypt",
    "django_bcrypt_sha256",
    "django_des_crypt",
    "django_disabled",
    "django_pbkdf2_sha1",
    "django_pbkdf2_sha256",
    "django_salted_md5",
    "django_salted_sha1",
    "grub_pbkdf2_sha512",
    "hex_digest",
    "ldap_crypt",
    "ldap_md5",
    "ldap_plaintext",
    "ldap_salted_md5",
    "ldap_salted_sha1",
    "ldap_salted_sha256",
    "ldap_salted_sha512",
    "ldap_sha1",
    "lmhash",
    "md5_crypt",
    "msdcc",
    "msdcc2",
    "mssql2000",
    "mssql2005",
    "mysql323",
    "mysql41",
    "nthash",
    "oracle10",
    "oracle11",
    "pbkdf2_sha1",
    "pbkdf2_sha256",
    "pbkdf2_sha512",
    "phpass",
    "plain",
    "plaintext",
    "postgres_md5",
    "scram",
    "scrypt",
    "sha256_crypt",
    "sha512_crypt",
]
ALL_ALGORITHMS: Tuple[TAlgorithm, ...] = get_args(TAlgorithm)



from rodi import Container as BaseContainer
from rodi import OverridingServiceException, ServiceLifeStyle

from .models import ServiceDefinitionCollection


class Container(BaseContainer):
    """Override's rodi container with our way to load services. This is a working implementation, although it would
    need polishing. Maybe the container should be reworked entirely to avoid the rodi duplications entirely, here the
    api methods using the builtin rodi providers/resolvers are still available, and that's maybe not what we want,
    for the long term."""

    def load(self, filename, *, bind_settings):
        """
        Loads a declarative service collection from a yaml file, and bind settings for config resolution.

        :param filename: str
        :param bind_settings: dict-like
        """
        from .resolvers import ServiceResolver

        collection = ServiceDefinitionCollection.model_validate_yaml(filename)

        if bind_settings:
            collection.bind_settings(bind_settings)

        for service in collection:
            lifestyle = getattr(ServiceLifeStyle, (service.lifestyle or "singleton").upper())
            resolver = ServiceResolver(self, service, lifestyle)

            if resolver.base_type in self._map:
                raise OverridingServiceException(resolver.base_type, resolver)
            self._map[resolver.base_type] = resolver

            if service.name in self._exact_aliases:
                raise OverridingServiceException(service.name, resolver.base_type)
            self.set_alias(service.name, resolver.base_type)



from pathlib import Path

import pytest
from hishel import AsyncCacheTransport
from rodi import CannotResolveTypeException
from whistle import AsyncEventDispatcher, IAsyncEventDispatcher

from harp import ROOT_DIR
from harp.services import Container
from harp.utils.testing.config import BaseConfigurableTest
from harp_apps.http_client.settings import CacheSettings, HttpClientSettings


class MyCustomTransport(AsyncCacheTransport):
    pass


class TestCacheSettings(BaseConfigurableTest):
    type = CacheSettings
    expected_verbose = {
        "controller": {
            "allow_heuristics": False,
            "allow_stale": False,
            "cacheable_methods": ["GET", "HEAD"],
            "cacheable_status_codes": [
                200,
                203,
                204,
                206,
                300,
                301,
                308,
                404,
                405,
                410,
                414,
                501,
            ],
            "type": "hishel.Controller",
        },
        "enabled": True,
        "storage": {
            "base": "hishel.AsyncBaseStorage",
            "check_ttl_every": 60.0,
            "ttl": None,
            "type": "harp_apps.http_client.contrib.hishel.storages.AsyncStorage",
        },
        "transport": {"type": "hishel.AsyncCacheTransport"},
    }

    def test_service_override(self):
        settings = HttpClientSettings(
            cache=self.create(
                transport={
                    "type": MyCustomTransport.__module__ + "." + MyCustomTransport.__qualname__,
                },
            )
        )

        container = Container()
        container.add_instance(AsyncEventDispatcher(), IAsyncEventDispatcher)
        container.load(
            Path(ROOT_DIR) / "harp_apps" / "http_client" / "services.yml",
            bind_settings=settings,
        )

        provider = container.build_provider()

        cache = provider.get("http_client.cache.transport")
        assert isinstance(cache, MyCustomTransport)

    def test_cache_disabled(self):
        settings = HttpClientSettings(cache=self.create(enabled=False))

        container = Container()
        container.add_instance(AsyncEventDispatcher(), IAsyncEventDispatcher)
        container.load(
            Path(ROOT_DIR) / "harp_apps" / "http_client" / "services.yml",
            bind_settings=settings,
        )

        provider = container.build_provider()

        with pytest.raises(CannotResolveTypeException):
            provider.get("http_client.cache.transport")



from .bridges import HttpRequestBridge, HttpResponseBridge
from .messages import BaseHttpMessage, BaseMessage
from .serializers import MessageSerializer

__all__ = [
    "BaseMessage",
    "BaseHttpMessage",
    "HttpRequestBridge",
    "HttpResponseBridge",
    "MessageSerializer",
]



import click
import pytest
from click import Context, UsageError

from harp.commandline.options.server import _EnhancedParserCommand
from harp.commandline.server import CommonServerOptions


def test_default():
    options = CommonServerOptions(options=(), files=(), enable=(), disable=())
    assert options.as_list() == []


def test_applications():
    options = CommonServerOptions(options=(), files=(), enable=("foo", "bar"), disable=("baz", "blurp"))
    assert options.as_list() == [
        "--enable foo",
        "--enable bar",
        "--disable baz",
        "--disable blurp",
    ]


def test_parse_args_default_behaviour():
    @click.command(cls=_EnhancedParserCommand)
    @click.option("--other")
    def cmd(): ...

    assert _call_command(cmd, "--other", "value") == {"other": "value"}
    with pytest.raises(UsageError):
        _call_command(cmd, "--other", "value", "unknown")


def test_parse_args_set_behaviour():
    @click.command(cls=_EnhancedParserCommand)
    @click.option("--set", "options", multiple=True, type=(str, str))
    def cmd(): ...

    assert _call_command(cmd, "--set", "arg=value", "--set", "foo", "bar") == {
        "options": (("arg", "value"), ("foo", "bar"))
    }


def test_parse_args_mixed_arguments_behaviour():
    @click.command(cls=_EnhancedParserCommand)
    @click.option("--foo")
    @click.option("--bar")
    @click.option("--set", "options", multiple=True, type=(str, str))
    @click.argument("positional", nargs=-1)
    def cmd(): ...

    assert _call_command(
        cmd,
        "--foo",
        "foo",
        "something",
        "--set",
        "arg=value",
        "completely",
        "--set",
        "foo",
        "bar",
        "different",
        "--bar",
        "bar",
    ) == {
        "foo": "foo",
        "bar": "bar",
        "options": (("arg", "value"), ("foo", "bar")),
        "positional": ("something", "completely", "different"),
    }


def _call_command(cmd, *args):
    ctx = Context(cmd)
    cmd.parse_args(ctx, list(args))
    return ctx.params



from harp.config import ConfigurationBuilder


class BaseRulesFlowTest:
    applications = []

    async def create_system(self, settings, /, *, mock):
        system = await ConfigurationBuilder(
            (settings or {}) | {"applications": self.applications},
            use_default_applications=False,
        ).abuild_system()
        system.config["rules"].ruleset.add({"*": {"*": {"*": [mock]}}})
        return system



from harp.controllers import GetHandler, RouterPrefix, RoutingController
from harp.http import HttpResponse
from harp_apps.storage.types import IBlobStorage


@RouterPrefix("/api/blobs")
class BlobsController(RoutingController):
    def __init__(self, *, storage: IBlobStorage, handle_errors=True, router=None):
        self.storage = storage
        super().__init__(handle_errors=handle_errors, router=router)

    @GetHandler("/{id}")
    async def get(self, id):
        blob = await self.storage.get(id)
        if not blob:
            return HttpResponse(b"Blob not found.", status=404, content_type="text/plain")

        content_type = blob.content_type or "application/octet-stream"

        try:
            data = blob.prettify()
        except (ValueError, NotImplementedError):
            data = blob.data

        return HttpResponse(data, content_type=content_type)



from functools import cached_property
from typing import override

from sqlalchemy import make_url
from sqlalchemy.exc import DatabaseError
from sqlalchemy.ext.asyncio import create_async_engine

from harp import get_logger

from ..utils.sql import run_sql
from ._base import BaseOptional

logger = get_logger(__name__)


class PgTrgmOptional(BaseOptional):
    """
    Installs or uninstalls the required tools for postrges' trigram based indexes and search.

    """

    def __init__(self, db_url):
        self._url = db_url

    @cached_property
    def engine(self):
        engine = create_async_engine(self.url)

        if engine.dialect.name != "postgresql":
            raise RuntimeError("This optional is only available for PostgreSQL databases.")

        return engine

    @cached_property
    def url(self):
        return make_url(self._url)

    @override
    async def is_supported(self):
        return self.url.get_dialect().name == "postgresql"

    @override
    async def install(self):
        """Install this optional feature in the database. This will install the pg_trgm extension and create optimized
        index for search."""
        is_pg_trm_available = await self._is_pg_trgm_extension_installed()

        if not is_pg_trm_available:
            try:
                await self._install_pg_trgm_extension_if_possible()
                is_pg_trm_available = await self._is_pg_trgm_extension_installed()
            except DatabaseError as e:
                logger.error(f"Failed to install pg_trgm extension: {e}")

        if not is_pg_trm_available:
            logger.error(
                "pg_trgm extension is not available, cannot create optimized indexes, will still create regular one."
            )
        else:
            await self._create_gin_index("messages", "summary")
            await self._create_gin_index("transactions", "endpoint")

    @override
    async def uninstall(self):
        """Remove this optional feature from database. Please note that the extension is not uninstalled, as it may
        have more side effect than we'd like too. Some deeper understanding of what is a PG extension and how scoped
        there are would be great."""
        await self._drop_index("messages", "summary")
        await self._drop_index("transactions", "endpoint")

        if await self._is_pg_trgm_extension_installed():
            logger.warning(
                "pg_trgm extension is installed and wont be removed, if you need to, you can remove it by hand."
            )

    async def _is_pg_trgm_extension_installed(self) -> bool:
        """Check if pg_trgm extension is already installed in the database."""
        result = await self._run_sql("SELECT 1 FROM pg_extension WHERE extname='pg_trgm'")
        return result.rowcount > 0

    async def _run_sql(self, sql):
        """Run a SQL statement on the database (via sqlalchemy DBAL engine)."""
        return await run_sql(self.engine, sql)

    async def _install_pg_trgm_extension_if_possible(self):
        """Make sure the pg_trgm extension is installed in the database. This may fail, we'll alert the user about
        that, although there may be good reasons for this to fail (no privilege ...)."""
        try:
            await self._run_sql("CREATE EXTENSION IF NOT EXISTS pg_trgm;")
        except DatabaseError as e:
            logger.error(f"Failed to install pg_trgm extension: {e}")

    async def _create_gin_index(self, table_name, column_name):
        """Create a GIN index on a column, using the pg_trgm extension."""
        await self._run_sql(
            f"CREATE INDEX {table_name}_{column_name}_gin ON {table_name} USING gin ({column_name} gin_trgm_ops);"
        )

    async def _drop_index(self, table_name, column_name):
        """Drop an index on a column."""
        await self._run_sql(f"DROP INDEX IF EXISTS {table_name}_{column_name}_gin;")



from typing import Callable, Optional

from httpx import Request, Response
from whistle import Event

from harp import get_logger

logger = get_logger(__name__)


#: Event fired when an external request is about to be sent by the HTTP client.
EVENT_FILTER_HTTP_CLIENT_REQUEST = "http_client.filter.request"

#: Event fired when an external response is received by the HTTP client.
EVENT_FILTER_HTTP_CLIENT_RESPONSE = "http_client.filter.response"


class HttpClientFilterEvent(Event):
    """
    Custom event class embedding an :class:`httpx.Request` and an optional :class:`httpx.Response`, allowing to filter
    external requests made by the client. It also allows script execution, providing some context for systems like
    filtering rules (see ``rules`` application).

    """

    def __init__(self, request: Request, response: Optional[Response] = None):
        self.request = request
        self.response = response

    def set_response(self, response: Response):
        if response is not None and not isinstance(response, Response):
            raise ValueError(
                f"Response must be an instance of httpx.Response, got {response!r} ({type(response).__module__}.{type(response).__qualname__})."
            )
        self.response = response

    @property
    def endpoint(self):
        return self.request.extensions.get("harp", {}).get("endpoint", "unknown")

    @property
    def fingerprint(self):
        return f"{self.request.method} {self.request.url.raw_path.decode()}"

    def _get_rule_name(self) -> str | None:
        if not self.name:
            return None
        return "on_remote_" + self.name.rsplit(".", 1)[-1]

    @property
    def criteria(self):
        return self.endpoint, self.fingerprint, self._get_rule_name()

    def create_execution_context(self):
        return {
            "logger": logger,
            "rule": self._get_rule_name(),
            "endpoint": self.endpoint,
            "request": self.request,
            "response": self.response,
            "stop_propagation": self.stop_propagation,
        }

    def execute_script(self, script: Callable):
        context = self.create_execution_context()
        script(context)
        if context["response"] != self.response:
            if context["response"] is not None:
                if not isinstance(context["response"], Response):
                    raise ValueError(
                        f"Response must be an instance of httpx.Response, got {context['response']!r} ({type(context['response']).__module__}.{type(context['response']).__qualname__})."
                    )
            self.set_response(context["response"])
        return context



from abc import ABC, abstractmethod


class BaseOptional(ABC):
    @abstractmethod
    async def is_supported(self) -> bool:
        """Is this optional supported in the current environment?"""
        raise NotImplementedError()

    @abstractmethod
    async def install(self):
        """Install this optional feature in the database."""
        raise NotImplementedError()

    @abstractmethod
    async def uninstall(self):
        """Uninstall this optional feature from the database."""
        raise NotImplementedError()



from operator import itemgetter

from sqlalchemy import text


async def run_sql(engine, sql, *, autocommit=True):
    if isinstance(sql, str):
        sql = text(sql)
    async with engine.connect() as conn:
        result = await conn.execute(sql)
        if autocommit:
            await conn.commit()
    return result


_get0 = itemgetter(0)


async def run_postgres_explain_analyze(engine, sql):
    return "\n".join(map(_get0, (await run_sql(engine, "EXPLAIN ANALYZE " + sql)).fetchall()))



DEFAULT_SYSTEM_CONFIG_FILENAMES = (
    "/etc/harp.yaml",
    "/etc/harp.yml",
)

DEFAULT_APPLICATIONS = (
    "http_client",
    "proxy",
    "storage",
    "dashboard",
    "harp_apps.contrib.sentry",
    "telemetry",
    "janitor",
    "rules",
)



from harp.commandline import entrypoint
from harp.utils.testing.cli import CliRunner


def test_run_command(httpbin, snapshot):
    runner = CliRunner()

    result = runner.invoke(entrypoint, ["rules", "run", "api=" + httpbin, "GET", "/hostname"])
    assert result.exit_code == 0
    assert result.output.strip() == snapshot


def test_lint_command(snapshot):
    runner = CliRunner()

    result = runner.invoke(entrypoint, ["rules", "lint"])
    assert result.exit_code == 0
    assert result.output.strip() == snapshot



from harp_apps.storage.utils.testing.sql import run_cli_migrate_command


async def test_migrations(database_url):
    result = await run_cli_migrate_command(database_url)
    if result.exit_code != 0:
        print(result.output)
    assert result.exit_code == 0



import importlib.util
import sys

from harp.commandline.options.server import server_command
from harp.commandline.server import CommonServerOptions
from harp.commandline.utils.manager import HARP_DASHBOARD_SERVICE
from harp.utils.commandline import click, code


def _get_service_name_for_humans(x: str):
    if ":" in x:
        return x.split(":", 1)[1]
    return x


def assert_package_is_available(package_name: str):
    if importlib.util.find_spec(package_name) is None:
        raise ModuleNotFoundError(f"Package {package_name!r} is not available.")


def assert_development_packages_are_available():
    assert_package_is_available("honcho")
    assert_package_is_available("watchfiles")


@server_command(
    short_help="Starts the local development environment.",
    help=f"""Starts the local development environment, using honcho to spawn a configurable set of processes that you
    can adapt to your needs. By default, it will starts the `dashboard` (frontend dev server) and `server` (python
    server) processes. For live instances, you'll prefer {code("harp server")}.""",
)
@click.option(
    "--with-docs/--no-docs",
    default=False,
    help="Append the sphinx doc process to the process list.",
)
@click.option(
    "--with-ui/--no-ui",
    default=False,
    help="Append the storybook process to the process list.",
)
@click.option(
    "--mock",
    is_flag=True,
    help="Enable mock data instead of real api data (dashboard only).",
)
@click.option(
    "--server-subprocess",
    "-XS",
    "server_subprocesses",
    metavar="NAME:PORT:CMD",
    multiple=True,
    help="Add a server subprocess to the list of services to start (experimental, can be used multiple times).",
)
@click.argument("services", nargs=-1)
def start(with_docs, with_ui, services, server_subprocesses, mock, **kwargs):
    try:
        assert_development_packages_are_available()
    except ModuleNotFoundError as exc:
        raise click.UsageError(
            "\n".join(
                (
                    "You must install development dependencies to start the development environment.",
                    "",
                    'Try to install the "dev" extra.',
                )
            )
        ) from exc

    from harp.commandline.utils.manager import (
        HARP_DOCS_SERVICE,
        HARP_UI_SERVICE,
        get_honcho_manager_factory_type,
        parse_server_subprocesses_options,
    )

    more_env = {}

    if not mock:
        more_env.setdefault(HARP_DASHBOARD_SERVICE, {})["DISABLE_MOCKS"] = "true"

    options = CommonServerOptions(**kwargs)
    _dashboard_devserver_port = options.options.get("dashboard.devserver.port", None)

    HonchoManagerFactory = get_honcho_manager_factory_type()
    manager_factory = HonchoManagerFactory(
        proxy_options=options.as_list(),
        dashboard_devserver_port=_dashboard_devserver_port,
    )

    services = {f"harp:{_name}" for _name in services or ()} or set(manager_factory.defaults)
    if with_docs or HARP_DOCS_SERVICE in services:
        services.add(HARP_DOCS_SERVICE)
    if with_ui or HARP_UI_SERVICE in services:
        services.add(HARP_UI_SERVICE)

    for _name, (_proxy_port, _path, _cmd, _port) in parse_server_subprocesses_options(server_subprocesses).items():
        if _name in services or _name in manager_factory.names:
            raise click.UsageError(f"Duplicate process name: {_name}.")
        services.add(_name)
        manager_factory.names.add(_name)
        manager_factory.proxy_ports[_name] = _proxy_port
        manager_factory.ports[_name] = _port
        manager_factory.commands[_name] = _cmd
        manager_factory.cwds[_name] = _path

    # allow to limit the services to start
    if services:
        unknown_services = set(services) - manager_factory.names
        if unknown_services:
            unknown_services_as_string = ", ".join(map(_get_service_name_for_humans, sorted(unknown_services)))
            known_services_as_string = ", ".join(map(_get_service_name_for_humans, sorted(manager_factory.names)))
            raise click.UsageError(
                f"Unknown services: {unknown_services_as_string}. Available: {known_services_as_string}."
            )

    manager = manager_factory.build(services, more_env=more_env)

    try:
        manager.loop()
    finally:
        manager.terminate()
        manager.kill()

    sys.exit(manager.returncode)



from typing import Optional

from harp.config import Configurable


class NotificationsSettings(Configurable):
    enabled: bool = True
    slack_webhook_url: Optional[str] = None
    google_chat_webhook_url: Optional[str] = None



from typing import Annotated, Literal, Optional, override

from pydantic import BaseModel, Field

from .base import BaseLiveness, BaseLivenessSettings, LivenessSubject


class NaiveLivenessSettings(BaseLivenessSettings):
    type: Literal["naive"]

    failure_threshold: Annotated[int, Field(gt=0)] = 1
    success_threshold: Annotated[int, Field(gt=0)] = 1

    def build_impl(self):
        return NaiveLiveness(settings=self)


class NaiveLivenessSubjectState(BaseModel):
    """Holds the internal state of the target subject. Will be attached as an attribute to the said subject, but only
    used by the liveness implementation."""

    failure_score: int = 0
    success_score: int = 0


class NaiveLiveness(BaseLiveness[NaiveLivenessSettings]):
    @classmethod
    def get_state_of(cls, subject: LivenessSubject) -> NaiveLivenessSubjectState:
        _attr = f"__{type(cls).__name__}__state__"
        if not hasattr(subject, _attr):
            setattr(subject, _attr, NaiveLivenessSubjectState())
        return getattr(subject, _attr)

    @override
    def success(self, subject: LivenessSubject) -> bool:
        """Returns a boolean indicating if a state change happened."""
        state = self.get_state_of(subject)

        state.success_score += 1
        state.failure_score = 0

        if state.success_score >= self.settings.success_threshold:
            return self.set_status_as_up_if_necessary(subject)

        return False

    @override
    def failure(self, subject: LivenessSubject, reason: Optional[str] = None) -> bool:
        """Returns a boolean indicating if a state change happened."""
        state = self.get_state_of(subject)

        state.success_score = 0
        state.failure_score += 1

        self.add_failure_reason(subject, reason)

        if state.failure_score >= self.settings.failure_threshold:
            return self.set_status_as_down_if_necessary(subject)

        return False



from structlog import get_logger

from harp.config import Application, OnBindEvent, OnReadyEvent
from harp_apps.proxy.events import EVENT_PROXY_ERROR, ProxyErrorEvent

from .settings import SentrySettings

logger = get_logger(__name__)


async def on_proxy_error(event: ProxyErrorEvent):
    import sentry_sdk

    sentry_sdk.capture_exception(event.error.exception)


async def on_bind(event: OnBindEvent):
    event.dispatcher.add_listener(EVENT_PROXY_ERROR, on_proxy_error)


async def on_ready(event: OnReadyEvent):
    settings = event.provider.get(SentrySettings)
    if not settings.dsn:
        return

    import sentry_sdk
    from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

    sentry_sdk.init(dsn=settings.dsn, traces_sample_rate=1.0)
    event.asgi_app = SentryAsgiMiddleware(event.asgi_app)


application = Application(
    on_bind=on_bind,
    on_ready=on_ready,
    settings_type=SentrySettings,
)



from sqlalchemy import ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from . import Base
from .base import Repository


class Tag(Base):
    __tablename__ = "tags"

    id = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)
    name = mapped_column(String(64), nullable=False, unique=True)


class TagsRepository(Repository[Tag]):
    Type = Tag


class TagValue(Base):
    __tablename__ = "tag_values"
    __table_args__ = (UniqueConstraint("tag_id", "value", name=f"_{__tablename__[3:]}_uc"),)

    id = mapped_column(Integer(), primary_key=True, unique=True, autoincrement=True)

    tag_id = mapped_column(ForeignKey("tags.id", ondelete="CASCADE"), nullable=False)
    tag: Mapped["Tag"] = relationship()

    value = mapped_column(String(255), nullable=False)


class TagValuesRepository(Repository[TagValue]):
    Type = TagValue



from datetime import UTC, datetime

from whistle import AsyncEventDispatcher

from harp.http import HttpRequest, HttpResponse
from harp.models import Transaction
from harp.utils.guids import generate_transaction_id_ksuid
from harp_apps.proxy.events import (
    EVENT_TRANSACTION_ENDED,
    EVENT_TRANSACTION_MESSAGE,
    EVENT_TRANSACTION_STARTED,
    HttpMessageEvent,
    TransactionEvent,
)
from harp_apps.storage.constants import (
    SKIP_REQUEST_BODY_STORAGE,
    SKIP_REQUEST_HEADERS_STORAGE,
    SKIP_REQUEST_STORAGE,
    SKIP_RESPONSE_BODY_STORAGE,
    SKIP_RESPONSE_HEADERS_STORAGE,
    SKIP_RESPONSE_STORAGE,
)
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.types import IBlobStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin
from harp_apps.storage.worker import StorageAsyncWorkerQueue


class TestStorageWorkerMarkers(StorageTestFixtureMixin):
    """Test storage worker behavior with body storage skip markers."""

    def create_worker(
        self,
        dispatcher: AsyncEventDispatcher,
        sql_storage: SqlStorage,
        blob_storage: IBlobStorage,
    ) -> StorageAsyncWorkerQueue:
        """Create and register a storage worker."""
        worker = StorageAsyncWorkerQueue(sql_storage.engine, sql_storage, blob_storage)
        worker.register_events(dispatcher)
        return worker

    async def create_transaction_and_messages(
        self,
        dispatcher: AsyncEventDispatcher,
        sql_storage: SqlStorage,
        blob_storage: IBlobStorage,
        markers: set[str] | None = None,
        request_body: bytes = b"request content",
        response_body: bytes = b"response content",
    ) -> Transaction:
        """Helper to create a transaction with request/response messages through events."""
        worker = self.create_worker(dispatcher, sql_storage, blob_storage)

        # Create transaction
        transaction = Transaction(
            id=generate_transaction_id_ksuid(),
            type="http",
            endpoint="/test",
            started_at=datetime.now(UTC),
        )
        if markers:
            transaction.markers = markers

        # Dispatch transaction started event
        await dispatcher.adispatch(EVENT_TRANSACTION_STARTED, TransactionEvent(transaction=transaction))

        # Create and dispatch request message event
        request = HttpRequest(method="POST", path="/test", body=request_body)
        request_event = HttpMessageEvent(transaction=transaction, message=request)
        await dispatcher.adispatch(EVENT_TRANSACTION_MESSAGE, request_event)

        # Create and dispatch response message event
        response = HttpResponse(response_body, status=200)
        response_event = HttpMessageEvent(transaction=transaction, message=response)
        await dispatcher.adispatch(EVENT_TRANSACTION_MESSAGE, response_event)

        # Mark transaction as finished
        transaction.finished_at = datetime.now(UTC)
        await dispatcher.adispatch(EVENT_TRANSACTION_ENDED, TransactionEvent(transaction=transaction))

        # Wait for worker to process all events
        await worker.wait_until_empty()

        return transaction

    async def test_normal_storage_without_markers(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that request and response bodies are stored normally when no markers are present."""
        dispatcher = AsyncEventDispatcher()

        transaction = await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]
        assert stored_transaction.id == transaction.id

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        assert request_msg.kind == "request"
        assert response_msg.kind == "response"

        # Verify both bodies were stored
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"

        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"response data"

    async def test_skip_request_body_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that request body is not stored when skip-request-body-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_BODY_STORAGE},
            request_body=b"sensitive request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        # Verify request body was NOT stored
        assert request_msg.kind == "request"
        assert request_msg.body is None

        # Verify request headers WERE stored
        assert request_msg.headers is not None

        # Verify response body WAS stored (marker only affects request)
        assert response_msg.kind == "response"
        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"response data"

    async def test_skip_response_body_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that response body is not stored when skip-response-body-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_RESPONSE_BODY_STORAGE},
            request_body=b"request data",
            response_body=b"sensitive response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        # Verify request body WAS stored (marker only affects response)
        assert request_msg.kind == "request"
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"

        # Verify response body was NOT stored
        assert response_msg.kind == "response"
        assert response_msg.body is None

        # Verify response headers WERE stored
        assert response_msg.headers is not None

    async def test_skip_both_request_and_response_body_storage(
        self, sql_storage: SqlStorage, blob_storage: IBlobStorage
    ):
        """Test that both request and response bodies are not stored when both markers are set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_BODY_STORAGE, SKIP_RESPONSE_BODY_STORAGE},
            request_body=b"sensitive request",
            response_body=b"sensitive response",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        # Verify neither body was stored
        assert request_msg.kind == "request"
        assert request_msg.body is None
        assert response_msg.kind == "response"
        assert response_msg.body is None

        # Verify headers WERE still stored for both
        assert request_msg.headers is not None
        assert response_msg.headers is not None

    async def test_headers_stored_when_body_skipped(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that headers are always stored even when body storage is skipped."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_BODY_STORAGE},
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        request_msg = stored_transaction.messages[0]

        # Verify headers were stored even though body was skipped
        assert request_msg.headers is not None
        headers_blob = await blob_storage.get(request_msg.headers)
        assert headers_blob.content_type == "http/headers"

        # Verify message metadata is complete
        assert request_msg.summary is not None
        assert request_msg.created_at is not None

    async def test_skip_request_headers_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that request headers are not stored when skip-request-headers-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_HEADERS_STORAGE},
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        # Verify request headers were NOT stored
        assert request_msg.kind == "request"
        assert request_msg.headers is None

        # Verify request body WAS stored
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"

        # Verify response headers WAS stored (marker only affects request)
        assert response_msg.kind == "response"
        assert response_msg.headers is not None

    async def test_skip_response_headers_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that response headers are not stored when skip-response-headers-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_RESPONSE_HEADERS_STORAGE},
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify messages were stored
        assert len(stored_transaction.messages) == 2
        request_msg = stored_transaction.messages[0]
        response_msg = stored_transaction.messages[1]

        # Verify request headers WAS stored (marker only affects response)
        assert request_msg.kind == "request"
        assert request_msg.headers is not None

        # Verify response headers were NOT stored
        assert response_msg.kind == "response"
        assert response_msg.headers is None

        # Verify response body WAS stored
        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"response data"

    async def test_skip_request_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that entire request message is not stored when skip-request-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_STORAGE},
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify only response message was stored
        assert len(stored_transaction.messages) == 1
        response_msg = stored_transaction.messages[0]

        # Verify it's the response message
        assert response_msg.kind == "response"
        assert response_msg.headers is not None
        assert response_msg.body is not None
        response_blob = await blob_storage.get(response_msg.body)
        assert response_blob.data == b"response data"

    async def test_skip_response_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that entire response message is not stored when skip-response-storage marker is set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_RESPONSE_STORAGE},
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify only request message was stored
        assert len(stored_transaction.messages) == 1
        request_msg = stored_transaction.messages[0]

        # Verify it's the request message
        assert request_msg.kind == "request"
        assert request_msg.headers is not None
        assert request_msg.body is not None
        request_blob = await blob_storage.get(request_msg.body)
        assert request_blob.data == b"request data"

    async def test_skip_both_request_and_response_storage(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        """Test that no messages are stored when both skip-request-storage and skip-response-storage markers are set."""
        dispatcher = AsyncEventDispatcher()

        await self.create_transaction_and_messages(
            dispatcher,
            sql_storage,
            blob_storage,
            markers={SKIP_REQUEST_STORAGE, SKIP_RESPONSE_STORAGE},
            request_body=b"request data",
            response_body=b"response data",
        )

        # Verify transaction was stored
        transactions = await sql_storage.get_transaction_list(username="anonymous", with_messages=True)
        assert len(transactions) == 1
        stored_transaction = transactions[0]

        # Verify no messages were stored
        assert len(stored_transaction.messages) == 0



from collections import OrderedDict
from typing import override

from sqlalchemy import delete, insert, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncEngine

from harp import get_logger
from harp.models import Blob
from harp_apps.storage.models import Blob as SqlBlob
from harp_apps.storage.types import IBlobStorage

logger = get_logger(__name__)


class LRUSet:
    def __init__(self, capacity):
        self.capacity = capacity
        self.items = OrderedDict()

    def add(self, item):
        if item in self.items:
            # Move to the end to mark as recently used
            self.items.move_to_end(item)
        else:
            self.items[item] = True
            if len(self.items) > self.capacity:
                # Remove the least recently used item
                self.items.popitem(last=False)

    def remove(self, item):
        if item in self.items:
            del self.items[item]

    def exists(self, item):
        return item in self.items

    def get_lru(self):
        if self.items:
            return next(iter(self.items))
        return None

    def __repr__(self):
        return f"LRUSet({set(self.items.keys())})"


class SqlBlobStorage(IBlobStorage):
    type = "sql"

    def __init__(self, engine: AsyncEngine):
        self.engine = engine
        self.seen = LRUSet(1000)

    @override
    async def get(self, blob_id):
        """
        Retrieve a blob from the database, using its hash.
        Returns None if not found.

        :param blob_id: sha1 hash of the blob
        :return: Blob or None
        """
        async with self.engine.connect() as conn:
            result = await conn.execute(
                select(SqlBlob).where(SqlBlob.id == blob_id),
            )
            row = result.fetchone()
        if row:
            return Blob(id=blob_id, data=row.data, content_type=row.content_type)

    @override
    async def put(self, blob: Blob) -> Blob:
        """
        Store a blob in the database.

        :param blob_id: sha1 hash of the blob
        :param data: blob data
        """
        if self.seen.exists(blob.id):
            return blob

        async with self.engine.connect() as conn:
            if not (
                await conn.execute(
                    select(
                        select(SqlBlob.id).where(SqlBlob.id == blob.id).exists(),
                    ),
                )
            ).scalar_one():
                try:
                    await conn.execute(
                        insert(SqlBlob).values(id=blob.id, data=blob.data, content_type=blob.content_type),
                    )
                    await conn.commit()
                except IntegrityError:
                    pass  # already there? that's fine!
        return blob

    @override
    async def force_put(self, blob: Blob) -> Blob:
        try:
            async with self.engine.connect() as conn:
                await conn.execute(
                    insert(SqlBlob).values(id=blob.id, data=blob.data, content_type=blob.content_type),
                )
                await conn.commit()
        except IntegrityError:
            async with self.engine.connect() as conn:
                await conn.execute(
                    update(SqlBlob).where(SqlBlob.id == blob.id).values(data=blob.data, content_type=blob.content_type),
                )
                await conn.commit()
        return blob

    @override
    async def delete(self, blob_id):
        """
        Delete a blob from the database.

        :param blob_id: sha1 hash of the blob
        """
        self.seen.remove(blob_id)
        async with self.engine.connect() as conn:
            await conn.execute(
                delete(SqlBlob).where(SqlBlob.id == blob_id),
            )
            await conn.commit()

    @override
    async def exists(self, blob_id: str) -> bool:
        """
        Check if a blob exists in the database.

        :param blob_id: sha1 hash of the blob
        :return: True if the blob exists, False otherwise
        """
        if self.seen.exists(blob_id):
            return True

        async with self.engine.connect() as conn:
            return bool(
                (
                    await conn.execute(
                        select(
                            select(SqlBlob.id).where(SqlBlob.id == blob_id).exists(),
                        ),
                    )
                ).scalar_one()
            )



import time
from typing import Literal, Optional, override

from pydantic import BaseModel

from harp_apps.proxy.settings.liveness.base import BaseLiveness, BaseLivenessSettings, LivenessSubject


class LeakyBucketLivenessSettings(BaseLivenessSettings):
    type: Literal["leaky"]

    #: The rate at which the bucket leaks (x per second).
    rate: float = 1.0

    #: The maximum capacity of the bucket.
    capacity: float = 60.0

    #: The threshold over which the remote is considered down.
    threshold: float = 10.0

    def build_impl(self):
        return LeakyBucketLiveness(settings=self)


class LeakyBucketLivenessSubjectState(BaseModel):
    last_checked: Optional[float] = None
    current: float = 0.0

    def leak(self, rate: float):
        if not self.last_checked:
            self.last_checked = time.time()
        current_time = time.time()
        elapsed_time = current_time - self.last_checked
        self.current = max(0.0, self.current - elapsed_time * rate)
        self.last_checked = current_time


class LeakyBucketLiveness(BaseLiveness[LeakyBucketLivenessSettings]):
    @classmethod
    def get_state_of(cls, subject: LivenessSubject) -> LeakyBucketLivenessSubjectState:
        _attr = f"__{type(cls).__name__}__state__"
        if not hasattr(subject, _attr):
            setattr(subject, _attr, LeakyBucketLivenessSubjectState())
        return getattr(subject, _attr)

    @override
    def success(self, subject: LivenessSubject) -> bool:
        state = self.get_state_of(subject)
        state.leak(self.settings.rate)

        if state.current < self.settings.threshold:
            return self.set_status_as_up_if_necessary(subject)

        return False

    @override
    def failure(self, subject: LivenessSubject, reason: Optional[str] = None) -> bool:
        state = self.get_state_of(subject)
        state.leak(self.settings.rate)
        state.current = min(self.settings.capacity, state.current + 1)

        self.add_failure_reason(subject, reason)

        if state.current >= self.settings.threshold:
            return self.set_status_as_down_if_necessary(subject)

        return False



from decimal import Decimal

import httpx
import pytest

from ..events import HttpClientFilterEvent


def valid_response_script(context):
    context["response"] = httpx.Response(200, text="Ok.")


def invalid_response_script(context):
    context["response"] = Decimal(200)


def none_response_script(context):
    context["response"] = None


def test_execution():
    event = HttpClientFilterEvent(httpx.Request("GET", "http://example.com"))

    event.execute_script(valid_response_script)
    assert event.response.status_code == 200

    with pytest.raises(ValueError):
        event.execute_script(invalid_response_script)
    assert event.response.status_code == 200

    event.execute_script(none_response_script)
    assert event.response is None



from unittest.mock import MagicMock, patch

from harp.config.utils import get_configuration_builder_type


@patch("os.environ.get")
@patch("builtins.__import__")
def test_get_configuration_builder_type_with_custom_edition(mock_import, mock_get_env):
    # Mock environment variable
    mock_get_env.return_value = "custom_edition"

    # Mock import of the custom edition
    mock_module = MagicMock()
    mock_import.return_value = mock_module
    mock_module.ConfigurationBuilder = "MockedCustomConfigurationBuilder"

    result = get_configuration_builder_type()
    assert result == "MockedCustomConfigurationBuilder"


def test_get_configuration_builder_type_fallback_to_default_edition():
    with patch.dict("os.environ", {"HARP_EDITION": "non_existent_edition"}):
        result = get_configuration_builder_type()
        assert result.__name__ == "ConfigurationBuilder"
        assert result.__module__ == "harp.config.builders.configuration"



from functools import cached_property
from typing import Optional

from multidict import MultiDict, MultiDictProxy

from .typing import BaseHttpMessage


class HttpError(BaseHttpMessage):
    kind = "error"

    def __init__(
        self,
        message: str,
        /,
        *,
        exception: Optional[Exception] = None,
        status=500,
        verbose_message=None,
    ):
        super().__init__()
        self.message = message
        self.exception = exception
        self.status = status
        self.verbose_message = verbose_message or message

    @cached_property
    def headers(self) -> MultiDictProxy:
        return MultiDictProxy(MultiDict())

    @cached_property
    def body(self) -> bytes:
        if self.exception:
            msg = str(self.exception)
            out = type(self.exception).__name__
            if msg:
                out += f": {msg}"
            return out.encode()
        return self.message.encode()



import operator
from typing import Any, Self

from pydantic import BaseModel

from harp.utils.config import yaml


class _NotSet:
    def __repr__(self):
        return "<not set>"


_notset = _NotSet()


class BaseReference(BaseModel):
    """
    Base class for references. Mostly there to add some interface documentation.

    """

    def __init__(self, /, **data: Any) -> None:  # type: ignore
        """Create a new model by parsing and validating input data from keyword arguments.

        Raises :class:`pydantic_core.ValidationError` if the input data cannot be validated.

        """
        super().__init__(**data)

    @classmethod
    def build_from_yaml(cls, loader, node) -> Self:
        """
        Secondary constructor implementing the pyyaml constructor interface.

        """
        raise NotImplementedError("YAML constructor not implemented")


class LazyServiceReference(BaseReference):
    """
    Reference to a service, that will be resolved the latest possible, when the instance will actually be needed.
    """

    #: Reference target, a.k.a. the service symbolic name. Can be either a string, or a list of strings. The later will
    #: resolve as the first available service in the list.
    target: str | list[str]

    @classmethod
    def build_from_yaml(cls, loader, node) -> Self:
        if isinstance(node, yaml.ScalarNode):
            return cls(target=loader.construct_scalar(node))

        elif isinstance(node, yaml.SequenceNode):
            values = loader.construct_sequence(node)
            return cls(target=values)

        raise ValueError(f"Unsupported node type: {type(node)}")

    def __repr__(self):
        return f"{type(self).__name__}({repr(self.target)})"

    def resolve(self, resolver, context):
        """
        Resolve reference value in using the given resolver (callable) and resolution context.
        """
        return resolver(self.target, context)


operators = {
    "==": operator.eq,
    " is not ": operator.is_not,
    " is ": operator.is_,
}


class LazySettingReference(BaseReference):
    """
    Reference to a setting value, that will be resolved when the settings are bound to the service definitions
    collection, when calling :meth:`harp.services.Container.load`.

    Use this in `services.yml` files using the `!cfg` constructor:

    .. code:: yaml

        foo: !cfg "foo"

    Can be provided a two-element list for defaults:

    .. code:: yaml

        foo: !cfg ["foo", "default if no «foo» in bound settings"]

    """

    target: str
    default: Any = _notset

    def __init__(self, target=None, **kwargs):
        super().__init__(target=target, **kwargs)

    @classmethod
    def build_from_yaml(cls, loader, node) -> Self:
        if isinstance(node, yaml.ScalarNode):
            return cls(loader.construct_scalar(node))
        elif isinstance(node, yaml.SequenceNode):
            values = loader.construct_sequence(node)
            assert len(values) == 2
            return cls(values[0], default=values[1])

        raise ValueError(f"Unsupported node type: {type(node)}")

    def resolve(self, settings):
        operator = None
        if "==" in self.target:
            target, operator, other_operand = self.target.partition("==")
        elif " is not " in self.target:
            target, operator, other_operand = self.target.partition(" is not ")
        elif " is " in self.target:
            target, operator, other_operand = self.target.partition(" is ")
        else:
            target = self.target
            other_operand = None

        x = settings
        for part in target.strip().split("."):
            try:
                x = x[part]
            except (TypeError, KeyError):
                x = self.default
                break

        if operator and other_operand:
            other_operand = eval(other_operand.strip())
            if operator in operators:
                return operators[operator](x, other_operand) or None
            else:
                raise ValueError(f"Unsupported operator: {operator}")

        return x if x is not _notset else None

    def __str__(self):
        return f"!cfg {self.target}"


yaml.add_constructor("!ref", LazyServiceReference.build_from_yaml, Loader=yaml.Loader)
yaml.add_constructor("!cfg", LazySettingReference.build_from_yaml, Loader=yaml.Loader)



from typing import Optional

from slack_sdk.webhook.async_client import AsyncWebhookClient


class SlackNotificationSender:
    def __init__(self, webhook_url: str, public_url: Optional[str] = None) -> None:
        self.webhook_url = webhook_url
        self.client = AsyncWebhookClient(webhook_url)
        self.public_url = public_url

    async def send_notification(
        self,
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
    ) -> None:
        slack_message = self._format_error_message(method, url, status_code, message, transaction_id, self.public_url)
        response = await self.client.send(text="ERROR NOTIFICATION", blocks=slack_message["blocks"])
        assert response.status_code == 200
        assert response.body == "ok"

    @staticmethod
    def _format_error_message(
        method: Optional[str],
        url: Optional[str],
        status_code: int,
        message: str,
        transaction_id: Optional[str],
        public_url: Optional[str] = None,
    ) -> dict:
        blocks = [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": ":red_circle: Error Notification",
                    "emoji": True,
                },
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"Error in {method} request to {url}",
                },
            },
            {
                "type": "section",
                "fields": [
                    {"type": "mrkdwn", "text": f"*Error Code:*\n`{status_code}`"},
                    {"type": "mrkdwn", "text": f"*Status:*\n{message}"},
                ],
            },
            {
                "type": "context",
                "elements": [
                    {
                        "type": "mrkdwn",
                        "text": f":pager: *Transaction ID:* `{transaction_id}`",
                    }
                ],
            },
        ]

        if public_url:
            blocks.append(
                {
                    "type": "actions",
                    "elements": [
                        {
                            "type": "button",
                            "text": {
                                "type": "plain_text",
                                "text": "View Logs",
                                "emoji": True,
                            },
                            "url": f"{public_url}/transactions/{transaction_id}",
                            "style": "primary",
                        },
                    ],
                },
            )
        return {"blocks": blocks}



from unittest import mock

from harp.config import ConfigurationBuilder
from harp.utils.testing.communicators import ASGICommunicator


async def test_notifications_with_proxy(httpbin):
    """
    Test that notifications are sent for 5xx errors from proxied requests.

    This test verifies that:
    1. The proxy correctly handles error responses (including 502)
    2. The notification system correctly extracts the reason_phrase from responses
    3. The notification is triggered with the correct status code and message

    Regression test for: reason_phrase property should handle both str and bytes values
    """
    settings = {
        "applications": ["http_client", "proxy", "notifications"],
        "proxy": {"endpoints": [{"name": "httpbin", "port": 80, "url": httpbin}]},
        "notifications": {
            "slack_webhook_url": "https://slack.com",
            "google_chat_webhook_url": "https://chat.google.com",
        },
    }

    system = await ConfigurationBuilder(settings, use_default_applications=False).abuild_system()

    client = ASGICommunicator(system.asgi_app)
    await client.asgi_lifespan_startup()

    with mock.patch("harp_apps.notifications.subscriber.NotificationSubscriber.send_notification") as send_notification:
        # Request a 502 status from httpbin
        await client.http_get("/status/502")

        # Verify notification was triggered
        assert send_notification.called, "send_notification should be called for 5xx errors"

        # Extract the call arguments
        _, send_notification_kwargs = send_notification.call_args

        # Verify the notification contains the correct information
        assert send_notification_kwargs["status_code"] == 502, "Status code should be 502"
        assert send_notification_kwargs["message"] == "Bad Gateway", "Message should be 'Bad Gateway'"
        assert send_notification_kwargs["method"] == "GET", "HTTP method should be GET"



from harp.utils.types import typeof


def test_typeof():
    assert typeof(1) == "builtins.int"
    assert typeof("1") == "builtins.str"
    assert typeof(1, short=True) == "int"
    assert typeof("1", short=True) == "str"
    assert typeof(object) == "builtins.type"



import math

from harp import get_logger
from harp.controllers import GetHandler, RouteHandler, RouterPrefix, RoutingController
from harp.http import HttpRequest, JsonHttpResponse
from harp.models.transactions import Transaction
from harp.settings import PAGE_SIZE
from harp.views.json import json
from harp_apps.storage.models.flags import FLAGS_BY_NAME
from harp_apps.storage.types import IStorage

from ..filters import (
    TransactionEndpointFacet,
    TransactionFlagFacet,
    TransactionMethodFacet,
    TransactionStatusFacet,
    TransactionTpdexFacet,
)

logger = get_logger(__name__)


@RouterPrefix("/api/transactions")
class TransactionsController(RoutingController):
    def __init__(self, *, storage: IStorage, handle_errors=True, router=None):
        self.storage = storage
        self.facets = {
            facet.name: facet
            for facet in (
                TransactionEndpointFacet(storage=self.storage),
                TransactionMethodFacet(),
                TransactionStatusFacet(),
                TransactionFlagFacet(),
                TransactionTpdexFacet(),
            )
        }

        super().__init__(handle_errors=handle_errors, router=router)

    @GetHandler("/filters")
    async def filters(self, request: HttpRequest):
        await self.facets["endpoint"].refresh()

        return json(
            {name: facet.filter_from_query(request.query) for name, facet in self.facets.items()},
        )

    @GetHandler("/")
    async def list(self, request: HttpRequest):
        page = int(request.query.get("page", 1))
        if page < 1:
            page = 1

        cursor = str(request.query.get("cursor", ""))

        results = await self.storage.get_transaction_list(
            with_messages=True,
            filters={name: facet.get_filter_from_query(request.query) for name, facet in self.facets.items()},
            page=page,
            cursor=cursor,
            username=request.extensions.get("user") or "anonymous",
            text_search=request.query.get("search", ""),
        )

        return json(
            {
                "items": list(map(Transaction.to_dict, results.items)),
                "pages": math.ceil(results.meta.get("total", 0) / PAGE_SIZE),
                "total": results.meta.get("total", 0),
                "perPage": PAGE_SIZE,
            }
        )

    @GetHandler("/{id}")
    async def get(self, request: HttpRequest, id):
        transaction = await self.storage.get_transaction(
            id,
            username=request.extensions.get("user") or "anonymous",
        )

        if not transaction:
            return JsonHttpResponse({"error": "Transaction not found"}, status=404)

        return JsonHttpResponse(transaction.to_dict())

    @RouteHandler("/{id}/flags/{flag}", methods=["PUT", "DELETE"])
    async def set_user_flag(self, request: HttpRequest, id, flag):
        username = request.extensions.get("user", None) or "anonymous"
        flag_id = FLAGS_BY_NAME.get(flag)

        await self.storage.set_user_flag(
            transaction_id=id,
            username=username,
            flag=flag_id,
            value=False if request.method == "DELETE" else True,
        )

        return JsonHttpResponse({"success": True})



from harp_apps.rules.models.rulesets import RuleSet


class RulesSettings:
    def __init__(self, /, **sources):
        self.ruleset = RuleSet()
        self.ruleset.add(sources)

    def _asdict(self, /, *, secure=True):
        return self.ruleset._asdict(secure=secure)

    def __repr__(self):
        return f"{self.__class__.__name__}(rules={self.ruleset})"

    def load(self, /, filename, *, prefix="rules"):
        if filename.endswith(".yaml") or filename.endswith(".yml"):
            from harp.utils.config import yaml

            self.ruleset.add(yaml.load(filename).get(prefix, {}))
        elif filename.endswith(".toml"):
            from harp.utils.config import toml

            self.ruleset.add(toml.load(filename).get(prefix, {}))
        else:
            raise ValueError(f"Unsupported file format: {filename}")



import glob
import inspect
import os.path
from os.path import dirname
from pkgutil import iter_modules

import pytest
from pydantic import BaseModel, ValidationError

from harp import ROOT_DIR
from harp.config import Stateful
from harp.config.asdict import asdict
from harp.utils.config import yaml
from harp.utils.packages import get_full_qualified_name


def list_submodules(module):
    try:
        modpath = module.__path__
    except AttributeError:
        # not a package, return empty list
        return []

    submodules = []
    for submodule in iter_modules(modpath):
        submodules.append(submodule.name)
    return submodules


all_apps = [
    ".".join(dirname(x).split(os.path.sep))
    for x in glob.glob("**/__app__.py", root_dir=ROOT_DIR, recursive=True)
    if "/cookiecutters/" not in x and not x.startswith("test-wheel/") and not x.startswith(".venv/")
]


def get_all_settings_classes(modname):
    modpath, modattr = modname.rsplit(".", 1)
    module = __import__(modname, fromlist=[modattr])
    for _name, _impl in inspect.getmembers(module, inspect.isclass):
        if not _impl.__module__.startswith(modpath):
            continue
        for submodname in list_submodules(module):
            yield from get_all_settings_classes(f"{modname}.{submodname}")
        yield _name, _impl


# Some settings classes requires some value to have a meaning. They are not usually the root settings but some
# subsettings that does not really make sense unless relevant configuration is passed. In that case, we provide the
# minimal set of settings required to instanciate it.
REQUIRED_SETTINGS = {
    "harp_apps.proxy.settings.endpoint.BaseEndpointSettings": {
        "name": "api",
        "port": 4000,
    },
    "harp_apps.proxy.settings.remote.endpoint.RemoteEndpointSettings": {
        "url": "https://www.example.com/",
    },
    "harp_apps.dashboard.settings.auth.User": {
        "password": "secret",
    },
}

REQUIRED_SETTINGS["harp_apps.proxy.settings.endpoint.EndpointSettings"] = REQUIRED_SETTINGS[
    "harp_apps.proxy.settings.endpoint.BaseEndpointSettings"
]
IGNORE_TYPES = {
    "harp_apps.proxy.settings.liveness.base.BaseLiveness",
}


@pytest.mark.parametrize("app", all_apps)
def test_all_applications_default_settings(app, snapshot):
    modname = app.rsplit(".", 1)[-1]
    submodules = list_submodules(__import__(app, fromlist=[modname]))
    if "settings" not in submodules:
        pytest.skip(f"No settings submodule found in {app}.")

    _types = _get_all_configurable_types_for_application(app)

    all_defaults = {}
    for _fullname, (_name, _type) in _types.items():
        _kwargs: dict = REQUIRED_SETTINGS.get(_fullname, {})

        if not issubclass(_type, BaseModel):
            continue

        if get_full_qualified_name(_type) in IGNORE_TYPES:
            continue

        if issubclass(_type, Stateful):
            _settings_type = _type.get_settings_type()
            _kwargs["settings"] = REQUIRED_SETTINGS.get(get_full_qualified_name(_settings_type), {})

        if _fullname in REQUIRED_SETTINGS:
            with pytest.raises(ValidationError):
                _type()

        instance = _type(**_kwargs)

        all_defaults[_fullname] = yaml.dump(asdict(instance))

    assert all_defaults == snapshot


def _get_all_configurable_types_for_application(app):
    all_settings_classes = {}
    for _name, _type in get_all_settings_classes(app + ".settings"):
        _fullname = f"{_type.__module__}.{_type.__qualname__}"
        if _fullname not in all_settings_classes:
            all_settings_classes[_fullname] = _name, _type
    return all_settings_classes



from typing import TYPE_CHECKING, Awaitable, Callable

from asgiref.typing import ASGIApplication
from whistle import Event

from harp.services import Container, Services
from harp.typing import GlobalSettings
from harp.utils.network import Bind

if TYPE_CHECKING:
    from harp.controllers import ProxyControllerResolver


class OnBindEvent(Event):
    """
    «Bind» event happens before the service container is resolved.

    It is the right place to define services that may not be resolvable yet (because their dependencies may, or may not,
    be defined already).

    """

    name = "harp.config.bind"

    #: Service container.
    container: Container

    #: Global settings.
    settings: GlobalSettings

    def __init__(self, container: Container, settings: GlobalSettings):
        self.container = container
        self.settings = settings


#: Event fired when the service container is being configured, before the dependencies are resolved. You can configure
#: any services in a listener to this event, but all dependencies must be resolvable once the event is fully dispatched.
EVENT_BIND = OnBindEvent.name

OnBindHandler = Callable[[OnBindEvent], Awaitable[None]]


class OnBoundEvent(Event):
    """
    «Bound» event happens after the service container is resolved, before the kernel creation.

    It is the right place to setup things that needs the container to be resolved (thus, you get a `provider` instead
    of a `container`).

    """

    name = "harp.config.bound"

    #: Services provider.
    provider: Services

    #: Proxy controller resolver.
    resolver: "ProxyControllerResolver"

    def __init__(self, provider: Services, resolver: "ProxyControllerResolver"):
        self.provider = provider
        self.resolver = resolver


#: Event fired when the service container is fully resolved, before the kernel is created. You can setup things that
#: require live instance of services here.
EVENT_BOUND = OnBoundEvent.name

OnBoundHandler = Callable[[OnBoundEvent], Awaitable[None]]


class OnReadyEvent(Event):
    """
    «Ready» event happens after the asgi kernel is created, just before the application starts.

    It is the right place to decorate the kernel with middlewares.

    """

    name = "harp.config.ready"

    def __init__(self, provider: Services, asgi_app: ASGIApplication, binds: list[Bind]):
        self.provider = provider
        self.asgi_app = asgi_app
        self.binds = binds


#: Event fired when the application is ready to start. You can decorate the kernel with middlewares here.
EVENT_READY = OnReadyEvent.name

OnReadyHandler = Callable[[OnReadyEvent], Awaitable[None]]


class OnShutdownEvent(Event):
    """
    «Shutdown» event happens when the application is stopping.

    """

    name = "harp.config.shutdown"

    def __init__(self, asgi_app: ASGIApplication, provider: Services):
        self.asgi_app = asgi_app
        self.provider = provider


#: Event fired when the application is shutting down. You can cleanup resources here.
EVENT_SHUTDOWN = OnShutdownEvent.name

OnShutdownHandler = Callable[[OnShutdownEvent], Awaitable[None]]



from typing import Union
from urllib.parse import urlparse, urlunparse

import httpcore
from hishel._utils import normalized_url as _normalize_url


def extract_hostname_from_url(url):
    return urlparse(url).hostname


def normalize_url(url: Union[httpcore.URL, str, bytes]) -> str:
    url = _normalize_url(url)

    parsed_url = urlparse(url)

    if not parsed_url.scheme or not parsed_url.netloc:
        raise ValueError("Invalid URL: Missing scheme or netloc.")

    return urlunparse(
        parsed_url._replace(
            scheme=parsed_url.scheme.lower(),
            netloc=parsed_url.netloc.lower(),
            path=parsed_url.path or "/",
        )
    )



"""make message headers nullable

Revision ID: 29d104241dae
Revises: 4bdd9e1d790d
Create Date: 2025-10-30 11:25:43.000000

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "29d104241dae"
down_revision: Union[str, None] = "4bdd9e1d790d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("messages", "headers", existing_type=sa.VARCHAR(length=40), nullable=True, existing_nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("messages", "headers", existing_type=sa.VARCHAR(length=40), nullable=False, existing_nullable=True)
    # ### end Alembic commands ###



from datetime import UTC, datetime
from typing import Optional

from multidict import MultiDictProxy


class BaseMessage:
    protocol: str
    kind: str
    created_at: datetime

    def __init__(self, *, extensions: Optional[dict] = None):
        self.created_at = datetime.now(UTC)

        self._extensions = extensions or {}

    @property
    def extensions(self) -> dict:
        return self._extensions

    async def aread(self):
        pass


class BaseHttpMessage(BaseMessage):
    protocol = "http"

    @property
    def headers(self) -> MultiDictProxy:
        raise NotImplementedError

    @property
    def body(self) -> bytes:
        raise NotImplementedError



from typing import AsyncIterator, Mapping
from urllib.parse import parse_qsl

from httpx import AsyncByteStream
from multidict import CIMultiDict, MultiDict

from harp.http import HttpRequestBridge


class AsyncByteStreamFromBody(AsyncByteStream):
    def __init__(self, body) -> None:
        self._body = body

    async def __aiter__(self) -> AsyncIterator[bytes]:
        for chunk in self._body:
            yield chunk

    async def aclose(self) -> None:
        pass


class HttpRequestStubBridge(HttpRequestBridge):
    def __init__(
        self,
        /,
        *,
        server_ipaddr: str = "127.0.0.1",
        server_port: int = 80,
        method: str = "GET",
        path: str = "/",
        query: Mapping | str = None,
        headers: Mapping = None,
        body: bytes | list | None = None,
    ):
        self._server_ipaddr = server_ipaddr
        self._server_port = server_port
        self._method = str(method).upper()
        self._path = path

        if isinstance(body, bytes):
            self._body = [body]
        elif isinstance(body, list):
            self._body = body
        else:
            self._body = []
        self._closed = False

        if isinstance(query, str):
            query = parse_qsl(query)
        self._query = MultiDict(query or {})
        self._headers = CIMultiDict(headers or {})

    def get_server_ipaddr(self) -> str:
        return self._server_ipaddr

    def get_server_port(self) -> int:
        return self._server_port

    def get_method(self) -> str:
        return self._method

    def get_path(self) -> str:
        return self._path

    def get_query(self) -> MultiDict:
        return self._query

    def get_headers(self) -> CIMultiDict:
        return self._headers

    def get_stream(self) -> AsyncByteStream:
        return AsyncByteStreamFromBody(self._body)

    async def aread(self) -> bytes:
        return b"".join(self._body)



from os.path import dirname
from pathlib import Path

from harp import get_logger
from harp.config import Application, OnBindEvent

from .settings import HttpClientSettings

logger = get_logger(__name__)


async def on_bind(event: OnBindEvent):
    # Load service definitions, bound to our settings.
    event.container.load(
        Path(dirname(__file__)) / "services.yml",
        bind_settings=event.settings["http_client"],
    )


application = Application(
    on_bind=on_bind,
    settings_type=HttpClientSettings,
)



from inspect import signature
from typing import Optional

from http_router import NotFoundError, Router
from http_router.types import TMethodsArg, TPath
from rich.console import Console
from rich.traceback import Traceback

from harp.http import HttpRequest, HttpResponse
from harp.meta import get_meta, has_meta, set_meta
from harp.utils.arguments import Arguments


def get_exception_traceback_str(exc):
    console = Console(record=True)
    traceback = Traceback.from_exception(type(exc), exc, exc.__traceback__)
    console.print(traceback)
    return console.export_html()


class RoutingController:
    RouterType = Router
    RouterArguments = Arguments(trim_last_slash=True)

    prefix = None

    def __init__(self, *, handle_errors=True, router=None):
        self.router = router or self.create_router()
        self.handle_errors = handle_errors
        self.prefix = self.prefix or ""

        for _name in dir(self):
            _attr = getattr(self, _name)
            if callable(_attr) and (meta := get_meta(_attr, "route")):
                _paths, _methods, _opts = meta
                self.router.route(
                    *tuple(map(lambda x: self.prefix + x, _paths)),
                    methods=_methods,
                    **_opts,
                )(_attr)

        self.configure()

    def configure(self):
        pass

    def create_router(self):
        return self.RouterType(*self.RouterArguments.args, *self.RouterArguments.kwargs)

    async def __call__(self, request: HttpRequest):
        try:
            match = self.router(request.path, method=request.method)
            sig = signature(match.target)
            args = (
                request if name == "request" else match.params.get(name, param.default)
                for name, param in sig.parameters.items()
            )
            return await match.target(*args)
        except NotFoundError as exc:
            if not self.handle_errors:
                raise
            return self.handle_error(exc, status=404)
        except Exception as exc:
            if not self.handle_errors:
                raise
            return self.handle_error(exc)

    def handle_error(self, exc, *, status=500):
        return HttpResponse(
            f"<h1>{type(exc).__name__}</h1><h2>Stack trace</h2>{get_exception_traceback_str(exc)}",
            status=status,
            content_type="text/html",
        )


def RouterPrefix(prefix):
    def _configurator(Controller: RoutingController):
        Controller.prefix = prefix
        return Controller

    return _configurator


def RouteHandler(*paths: TPath, methods: Optional[TMethodsArg] = None, **opts):
    def _configurator(handler):
        nonlocal paths, methods, opts
        if has_meta(handler, "route"):
            raise Exception("Handler already has route metadata.")
        set_meta(handler, "route", (paths, methods, opts))
        return handler

    return _configurator


def GetHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["GET"], **opts)


def PostHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["POST"], **opts)


def PutHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["PUT"], **opts)


def DeleteHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["DELETE"], **opts)


def PatchHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["PATCH"], **opts)


def OptionsHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["OPTIONS"], **opts)


def HeadHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["HEAD"], **opts)


def TraceHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["TRACE"], **opts)


def ConnectHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=["CONNECT"], **opts)


def AnyMethodHandler(*paths: TPath, **opts):
    return RouteHandler(*paths, methods=None, **opts)



from httpx import AsyncClient

from harp.controllers import ProxyControllerResolver
from harp.http import HttpRequest
from harp_apps.proxy.controllers import HttpProxyController
from harp_apps.proxy.settings.endpoint import Endpoint


def test_add():
    http_client = AsyncClient()
    endpoint = Endpoint.from_kwargs(settings={"name": "test-endpoint", "port": 8080, "url": "http://example.com/"})
    resolver = ProxyControllerResolver()
    controller = HttpProxyController(http_client=http_client, remote=endpoint.remote, name=endpoint.settings.name)

    resolver.add_endpoint(endpoint, controller=controller)
    assert resolver.endpoints["test-endpoint"] == endpoint
    assert resolver.ports == (8080,)


async def test_resolve():
    http_client = AsyncClient()
    endpoint = Endpoint.from_kwargs(settings={"name": "test-endpoint", "port": 8080, "url": "http://example.com/"})
    resolver = ProxyControllerResolver()
    controller = HttpProxyController(http_client=http_client, remote=endpoint.remote, name=endpoint.settings.name)
    resolver.add_endpoint(endpoint, controller=controller)

    request = HttpRequest(server_port=8080)
    controller = await resolver.resolve(request)
    assert isinstance(controller, HttpProxyController)
    assert controller.name == "test-endpoint"

    request = HttpRequest(server_port=8081)
    controller = await resolver.resolve(request)
    assert controller is resolver.default_controller


def test_resolve_from_endpoint_name():
    http_client = AsyncClient()
    endpoint = Endpoint.from_kwargs(settings={"name": "test-endpoint", "port": 8080, "url": "http://example.com/"})
    resolver = ProxyControllerResolver()
    controller = HttpProxyController(http_client=http_client, remote=endpoint.remote, name=endpoint.settings.name)
    resolver.add_endpoint(endpoint, controller=controller)

    controller = resolver.resolve_by_name("test-endpoint")
    assert isinstance(controller, HttpProxyController)
    assert controller.name == "test-endpoint"

    controller = resolver.resolve_by_name("non-existent")
    assert controller is resolver.default_controller



from typing import Any, Generic, TypeVar

from dataclasses_jsonschema import JsonSchemaMixin


class Entity(JsonSchemaMixin):
    id: Any


TResult = TypeVar("TResult")


class Results(Generic[TResult]):
    def __init__(self):
        self.items: list[TResult] = []
        self.meta = {}

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item):
        return self.items[item]

    def append(self, item: TResult):
        self.items.append(item)



"""
Notifications Application
"""

from harp import get_logger
from harp.config import Application
from harp.config.events import OnBoundEvent, OnShutdownEvent
from harp.services import CannotResolveTypeException
from harp_apps.dashboard.settings import DashboardSettings
from harp_apps.notifications.settings import NotificationsSettings
from harp_apps.notifications.subscriber import NotificationSubscriber

logger = get_logger(__name__)

NOTIFICATIONS_SUBSCRIBER = "notifications.subscriber"


async def on_bound(event: OnBoundEvent):
    notifications_settings = event.provider.get(NotificationsSettings)

    public_url = None
    try:
        dashboard_settings = event.provider.get(DashboardSettings)
        public_url = dashboard_settings.public_url
    except CannotResolveTypeException:
        pass

    subscriber = NotificationSubscriber(notifications_settings, public_url)
    subscriber.subscribe(event.dispatcher)
    event.provider.set(NOTIFICATIONS_SUBSCRIBER, subscriber)


async def on_shutdown(event: OnShutdownEvent):
    subscriber = event.provider.get(NOTIFICATIONS_SUBSCRIBER)
    subscriber.unsubscribe(event.dispatcher)


application = Application(
    on_bound=on_bound,
    on_shutdown=on_shutdown,
    settings_type=NotificationsSettings,
)



class Arguments:
    __slots__ = ["_args", "_kwargs"]

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs



import os
from functools import lru_cache
from glob import glob
from os.path import dirname

import harp
import harp_apps

EXTENSIONS = [".yml", ".yaml", ".toml"]

# TODO this is clearly not the right place, and it's too static for the future, but let's keep it simple first.
APP_NAMESPACES = ["harp_apps", "harp_contrib"]


@lru_cache
def get_examples_dirname(*, app=None):
    if app is None:
        return os.path.dirname(os.path.abspath(__file__))
    for namespace in APP_NAMESPACES:
        dirname = os.path.join(harp.ROOT_DIR, namespace, app, "examples")
        if os.path.isdir(dirname):
            return dirname
    raise FileNotFoundError(f"Examples directory not found for app {app}.")


def guess_extension(basename):
    if basename.endswith(tuple(EXTENSIONS)):
        return basename
    for ext in EXTENSIONS:
        if os.path.isfile(basename + ext):
            return basename + ext
    raise FileNotFoundError(
        f"Example not found (looked for {basename}.{{{','.join(map(lambda x: x.lstrip('.'), EXTENSIONS))}}})"
    )


def get_example_filename(name):
    if ":" in name:
        app, name = name.split(":", 1)
        return guess_extension(os.path.join(get_examples_dirname(app=app), name))
    return guess_extension(os.path.join(get_examples_dirname(), name))


def get_available_examples_namespaces():
    _default_examples = dirname(__import__("harp.config.examples", fromlist=["__file__"]).__file__)

    _namespaces = {
        None: _default_examples,
    }

    _apps_root_dir = dirname(harp_apps.__file__)

    for _dirname in glob("**/examples", root_dir=_apps_root_dir, recursive=True):
        _pkg, _ = _dirname.rsplit("/", 1)
        _pkg = ".".join(("harp_apps", *_pkg.split("/")))
        _app_pkg = _pkg + ".__app__"
        try:
            __import__(_app_pkg)
        except ImportError:
            continue
        try:
            _example_dirname = dirname(__import__(_pkg + ".examples", fromlist=["__file__"]).__file__)
        except (ImportError, AttributeError):
            continue

        _namespace = _pkg.rsplit(".")[-1]
        if os.path.isdir(_example_dirname):
            _namespaces[_namespace] = _example_dirname
    return _namespaces


@lru_cache
def get_available_examples():
    examples = []
    for _namespace, _dirname in get_available_examples_namespaces().items():
        examples.extend(
            (
                ":".join(filter(None, (_namespace, os.path.splitext(os.path.basename(x))[0])))
                for x in glob(_dirname + "/*.yml")
            )
        )
    return list(sorted(examples))



import asyncio

from harp import get_logger
from harp.config import Application
from harp.config.events import OnBindEvent, OnBoundEvent, OnShutdownEvent
from harp_apps.janitor.worker import JanitorWorker
from harp_apps.storage.types import IStorage

logger = get_logger(__name__)

JANITOR_WORKER_TASK = "janitor.worker.task"


async def on_bind(event: OnBindEvent):
    event.container.add_singleton(JanitorWorker)


async def on_bound(event: OnBoundEvent):
    await event.provider.get(IStorage).ready()

    worker = event.provider.get(JanitorWorker)
    event.provider.set(JANITOR_WORKER_TASK, asyncio.create_task(worker.run()))


async def on_shutdown(event: OnShutdownEvent):
    worker = event.provider.get(JanitorWorker)
    worker.stop()
    event.provider.get(JANITOR_WORKER_TASK).cancel()


application = Application(
    on_bind=on_bind,
    on_bound=on_bound,
    on_shutdown=on_shutdown,
    dependencies=["storage"],
)



from rodi import (
    CannotResolveParameterException,
    CannotResolveTypeException,
    CircularDependencyException,
    DIException,
    FactoryMissingContextException,
    MissingTypeException,
    OverridingServiceException,
    Services,
)

from .containers import Container
from .references import LazyServiceReference

__all__ = [
    "CannotResolveParameterException",
    "CannotResolveTypeException",
    "CircularDependencyException",
    "Container",
    "DIException",
    "FactoryMissingContextException",
    "MissingTypeException",
    "OverridingServiceException",
    "LazyServiceReference",
    "Services",
]



from harp.models import Blob
from harp_apps.storage.types import IBlobStorage


class MemoryBlobStorage(IBlobStorage):
    type = "memory"

    def __init__(self):
        super().__init__()
        self._blobs = {}

    async def get(self, blob_id: str):
        return self._blobs.get(blob_id, None)

    async def put(self, blob: Blob) -> Blob:
        self._blobs[blob.id] = blob
        return blob

    force_put = put

    async def delete(self, blob_id: str):
        del self._blobs[blob_id]

    async def exists(self, blob_id: str) -> bool:
        return False



import os


def get_configuration_builder_type():
    edition = os.environ.get("HARP_EDITION", "harp_apps")
    try:
        ConfigurationBuilder = __import__(edition, fromlist=["ConfigurationBuilder"]).ConfigurationBuilder
    except (ImportError, AttributeError):
        from harp.config import ConfigurationBuilder
    return ConfigurationBuilder



from typing import cast

from click import Command

from harp import get_logger
from harp.utils.commandline import click

from .lint import lint_command
from .run import run_command

logger = get_logger(__name__)


@click.group("rules")
def entrypoint():
    """Rules engine related commands."""
    pass


entrypoint.add_command(cast(Command, lint_command))
entrypoint.add_command(cast(Command, run_command))



from .transaction_endpoint import TransactionEndpointFacet
from .transaction_flag import TransactionFlagFacet
from .transaction_method import TransactionMethodFacet
from .transaction_status import TransactionStatusFacet
from .transaction_tpdex import TransactionTpdexFacet
from .utils import flatten_facet_value

__all__ = [
    "TransactionMethodFacet",
    "TransactionStatusFacet",
    "TransactionEndpointFacet",
    "TransactionFlagFacet",
    "TransactionTpdexFacet",
    "flatten_facet_value",
]



import httpx
import pytest
import respx

from harp.http import HttpRequest
from harp_apps.proxy.constants import DOWN, UP
from harp_apps.proxy.controllers import HttpProxyController
from harp_apps.proxy.settings.remote import Remote

BASE_URL = "http://example.com"


@respx.mock
@pytest.mark.parametrize("liveness_type", [None, "ignore", "naive", "inherit", "leaky"])
async def test_basic(liveness_type):
    """
    This test ensures that the controller can make a request to a remote server and return the response. It's the
    baseline asumption for all other tests. This should be true for any liveness implementation.
    """
    respx.get(BASE_URL).mock(return_value=httpx.Response(200))

    remote = Remote.from_settings_dict(
        {
            **({"liveness": {"type": liveness_type}} if liveness_type is not None else {}),
            "endpoints": [{"url": BASE_URL}],
        }
    )
    controller = HttpProxyController(remote, http_client=httpx.AsyncClient())
    response = await controller(HttpRequest())

    assert response.status == 200
    assert remote[BASE_URL].status == UP


@respx.mock
@pytest.mark.parametrize("status", [500, 502, 503, 504])
@pytest.mark.parametrize("is_http_server_error_a_down_signal", [False, True])
async def test_break_on_5xx(status, is_http_server_error_a_down_signal):
    """
    This test ensures that the circuit breaker is triggered (opened) when a 5xx error is returned by the server, using
    the default settings. After opening, the next request should return a 503 status code.

    This requires a liveness implementation to be configured, as the default is now to "ignore".

    """
    respx.get(BASE_URL).mock(return_value=httpx.Response(status))

    remote = Remote.from_settings_dict(
        {
            "liveness": {"type": "naive"},
            "endpoints": [{"url": BASE_URL}],
            "break_on": (["http_5xx", "network_error"] if is_http_server_error_a_down_signal else ["network_error"]),
        }
    )
    endpoint = remote[BASE_URL]

    controller = HttpProxyController(remote, http_client=httpx.AsyncClient())
    initial_status = endpoint.status

    response = await controller(HttpRequest())
    if is_http_server_error_a_down_signal:
        assert response.status == status
        assert endpoint.status == DOWN
    else:
        assert response.status == status
        assert endpoint.status == initial_status

    response = await controller(HttpRequest())
    if is_http_server_error_a_down_signal:
        assert response.status == 503
        assert endpoint.status == DOWN
    else:
        assert response.status == status
        assert endpoint.status == initial_status


@respx.mock
@pytest.mark.parametrize("status", [400, 401, 403, 404])
@pytest.mark.parametrize("is_client_error_a_down_signal", [False, True])
async def test_break_on_4xx_if_configured(status, is_client_error_a_down_signal):
    """
    This test ensures that the circuit breaker is triggered (opened) when a 4xx error is returned by the server, if and
    only if the "break_on" setting contains "http_4xx". It requires a liveness alorithm to be configured.
    """

    respx.get(BASE_URL).mock(return_value=httpx.Response(status))

    remote = Remote.from_settings_dict(
        {
            "liveness": {"type": "naive"},
            "endpoints": [{"url": BASE_URL}],
            "break_on": (["http_4xx", "network_error"] if is_client_error_a_down_signal else ["network_error"]),
        }
    )
    endpoint = remote[BASE_URL]

    controller = HttpProxyController(remote, http_client=httpx.AsyncClient())
    initial_status = endpoint.status

    response = await controller(HttpRequest())
    if is_client_error_a_down_signal:
        assert response.status == status
        assert endpoint.status == DOWN
    else:
        assert response.status == status
        assert endpoint.status == initial_status

    response = await controller(HttpRequest())
    if is_client_error_a_down_signal:
        assert response.status == 503
        assert endpoint.status == DOWN
    else:
        assert response.status == status
        assert endpoint.status == initial_status


@respx.mock
@pytest.mark.parametrize("status", [400, 401, 403, 404, 500, 502, 503, 504])
async def test_do_not_break_if_break_on_explicitely_set(status):
    """
    This test ensures that the circuit breaker is not triggered when configured to break on nothing. It indirectly
    ensures that the break on setting is actually used.
    """
    respx.get(BASE_URL).mock(return_value=httpx.Response(status))

    remote = Remote.from_settings_dict(
        {
            "endpoints": [{"url": BASE_URL}],
            "break_on": [],
        }
    )
    endpoint = remote[BASE_URL]

    controller = HttpProxyController(remote, http_client=httpx.AsyncClient())
    initial_status = endpoint.status

    response = await controller(HttpRequest())
    assert response.status == status
    assert endpoint.status == initial_status

    response = await controller(HttpRequest())
    assert response.status == status
    assert endpoint.status == initial_status


@respx.mock
@pytest.mark.parametrize(
    ["error", "error_status"],
    [
        [httpx.NetworkError("oups"), 503],
        [httpx.TimeoutException("oups"), 504],
        [httpx.RemoteProtocolError("oups"), 502],
    ],
)
async def test_break_on_network_error(error, error_status):
    """
    This test ensures that the circuit breaker is triggered (opened) when a network error is raised (a network (L4)
    error is an error happening before we're actually talking HTTP (L7)). It requires a liveness alorithm to be
    selected, as the default "ignore" algorithm won't do anything.
    """
    respx.get(BASE_URL).mock(side_effect=error)

    remote = Remote.from_settings_dict(
        {
            "liveness": {"type": "naive"},
            "endpoints": [{"url": BASE_URL}],
        }
    )
    endpoint = remote[BASE_URL]

    controller = HttpProxyController(remote, http_client=httpx.AsyncClient())

    response = await controller(HttpRequest())
    assert response.status == error_status
    assert endpoint.status == DOWN

    response = await controller(HttpRequest())
    assert response.status == 503
    assert endpoint.status == DOWN



from unittest.mock import ANY

from sqlalchemy import text

from harp.models import Blob
from harp_apps.storage.services.blob_storages.sql import SqlBlobStorage


async def test_basics(sql_blob_storage: SqlBlobStorage):
    _storage = sql_blob_storage
    blob = Blob(id="foo", data=b"bar", content_type="text/plain")

    assert _storage.type == "sql"
    assert await _storage.get("foo") is None
    async with _storage.engine.connect() as conn:
        assert (await conn.execute(text("SELECT * FROM blobs WHERE id = 'foo'"))).fetchone() is None

    assert await _storage.put(blob) == blob
    assert await _storage.get("foo") == blob
    async with _storage.engine.connect() as conn:
        assert (
            await conn.execute(
                text("SELECT * FROM blobs WHERE id = 'foo'"),
            )
        ).fetchone() == ("foo", b"bar", "text/plain", ANY)

    assert await _storage.exists("foo")
    assert await _storage.delete("foo") is None
    assert not await _storage.exists("foo")
    assert await _storage.get("foo") is None
    async with _storage.engine.connect() as conn:
        assert (await conn.execute(text("SELECT * FROM blobs WHERE id = 'foo'"))).fetchone() is None



from typing import Generic, Type, TypeVar

from harp.config.asdict import asdict
from harp.config.configurables.base import BaseConfigurable

TConfigurable = TypeVar("TConfigurable", bound=BaseConfigurable)


class BaseConfigurableTest(Generic[TConfigurable]):
    type: Type[TConfigurable]
    initial = {}
    expected = {}
    expected_verbose = {}

    def create(self, **kwargs) -> TConfigurable:
        return self.type.from_kwargs(**{**self.initial, **kwargs})

    def test_defaults(self):
        obj = self.create()
        assert asdict(obj) == self.expected

    def test_defaults_verbose(self):
        obj = self.create()
        assert asdict(obj, verbose=True) == self.expected_verbose

    def test_jsonschema_for_validation(self, snapshot):
        schema = self.type.model_json_schema(mode="validation")
        schema.pop("description", None)
        assert schema == snapshot

    def test_jsonschema_for_serialization(self, snapshot):
        schema = self.type.model_json_schema(mode="serialization")
        schema.pop("description", None)
        assert schema == snapshot



"""
Proxy Application

"""

import asyncio
from asyncio import TaskGroup
from typing import cast

from harp.config import Application
from harp.config.events import OnBindEvent, OnBoundEvent, OnShutdownEvent
from harp.services.resolvers import ServiceResolver
from harp.utils.services import factory

from .settings import Proxy, ProxySettings

PROXY_HEALTHCHECKS_TASK = "proxy.healthchecks"


@factory(Proxy)
def ProxyFactory(self, settings: ProxySettings) -> Proxy:
    return Proxy(settings=settings)


async def create_background_task_group(coroutines):
    async def _execute():
        async with TaskGroup() as task_group:
            for coroutine in coroutines:
                task_group.create_task(coroutine)

    return asyncio.create_task(_execute())


async def on_bind(event: OnBindEvent):
    settings = event.settings.get("proxy")
    # add a controller service instance for each endpoint
    for endpoint in settings.endpoints:
        if endpoint.controller is not None:
            controller = endpoint.controller
            name = f"proxy.controllers.{endpoint.name}_controller"
            resolver = ServiceResolver(event.container, controller.to_service_definition(name, lifestyle="singleton"))
            event.container._map[name] = resolver

    event.container.add_singleton(Proxy, cast(type, ProxyFactory))


async def on_bound(event: OnBoundEvent):
    proxy: Proxy = event.provider.get(Proxy)

    for endpoint in proxy.endpoints:
        name = endpoint.settings.name
        event.provider._map[f"proxy.controllers.{name}_controller"].bind(remote=endpoint.remote, name=name)
        controller = event.provider.get(f"proxy.controllers.{name}_controller")
        event.resolver.add_endpoint(endpoint, controller=controller)

    event.provider.set(
        PROXY_HEALTHCHECKS_TASK,
        await create_background_task_group(
            [
                endpoint.remote.check_forever()
                for endpoint in proxy.endpoints
                if (endpoint.remote and endpoint.remote.probe)
            ]
        ),
    )


async def on_shutdown(event: OnShutdownEvent):
    await event.provider.get(PROXY_HEALTHCHECKS_TASK)._abort()


application = Application(
    dependencies=["services"],
    on_bind=on_bind,
    on_bound=on_bound,
    settings_type=ProxySettings,
)



from inspect import signature

import pytest
from rodi import ActivationScope, ServiceLifeStyle, Services

from harp.services.providers import ServiceProvider, filter_kwargs_based_on_signature


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        ({"a": 1, "b": 2, "c": 3}, {"a": 1, "b": 2}),
        ({"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 1, "b": 2}),
        ({"a": 1, "b": 2}, {"a": 1, "b": 2}),
        ({"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}, {"a": 1, "b": 2}),
        ({}, {}),
        ({"c": 3, "d": 4, "e": 5}, {}),
    ],
)
def test_filter_kwargs_based_on_signature(kwargs, expected):
    def func(a, b):
        pass

    sig = signature(func)
    assert filter_kwargs_based_on_signature(kwargs, sig) == expected


class TestServiceProvider:
    @pytest.fixture(autouse=True)
    def setup(self):
        class DummyService:
            def __init__(self, a, b):
                self.a = a
                self.b = b

            def constructor(self):
                self.a = 1
                self.b = 2

        self.args = []
        self.kwargs = {}
        self.constructor = DummyService
        self.lifestyle = ServiceLifeStyle.SINGLETON
        self.provider = ServiceProvider(self.constructor, args=self.args, kwargs=self.kwargs, lifestyle=self.lifestyle)

    def test_constructor_property(self):
        assert self.provider.constructor == self.constructor

        provider_with_constructor = ServiceProvider(
            self.constructor, _constructor="constructor", args=self.args, kwargs=self.kwargs, lifestyle=self.lifestyle
        )
        assert provider_with_constructor.constructor == self.constructor.constructor

    def test_signature_property(self):
        assert self.provider.signature == signature(self.constructor)

        provider_with_constructor = ServiceProvider(
            self.constructor, _constructor="constructor", args=self.args, kwargs=self.kwargs, lifestyle=self.lifestyle
        )
        assert provider_with_constructor.signature == signature(self.constructor.constructor)

    def test_call_method_missing_arguments(self):
        services = Services({"dummy_service": self.provider})
        activation_scope = ActivationScope(services)

        # without any args or kwargs specified this should throw an error
        with pytest.raises(TypeError):
            self.provider(activation_scope)

    def test_call_method_with_kwargs(self):
        provider = ServiceProvider(self.constructor, args=[], kwargs={"a": 1, "b": 2}, lifestyle=self.lifestyle)
        services = Services({"dummy_service": provider})
        activation_scope = ActivationScope(services)

        instance = provider(activation_scope)
        assert instance.a == 1
        assert instance.b == 2

    def test_call_method_with_args(self):
        provider = ServiceProvider(self.constructor, args=[1, 2], kwargs={}, lifestyle=self.lifestyle)
        services = Services({"dummy_service": provider})
        activation_scope = ActivationScope(services)

        instance = provider(activation_scope)
        assert instance.a == 1
        assert instance.b == 2

    def test_call_method_with_args_and_kwargs(self):
        provider = ServiceProvider(self.constructor, args=[1], kwargs={"b": 2}, lifestyle=self.lifestyle)
        services = Services({"dummy_service": provider})
        activation_scope = ActivationScope(services)

        instance = provider(activation_scope)
        assert instance.a == 1
        assert instance.b == 2

    def test_call_method_with_overlapping_args_and_kwargs(self):
        provider = ServiceProvider(self.constructor, args=[1], kwargs={"a": 1, "b": 2}, lifestyle=self.lifestyle)
        services = Services({"dummy_service": provider})
        activation_scope = ActivationScope(services)

        with pytest.raises(TypeError):
            provider(activation_scope)

    def test_bind_arguments(self):
        provider = ServiceProvider(self.constructor, args=[], kwargs={"a": 1}, lifestyle=self.lifestyle)
        services = Services({"dummy_service": provider})
        activation_scope = ActivationScope(services)

        with pytest.raises(TypeError):
            provider(activation_scope)

        # now binding the missing argument
        provider.bind(b=2)

        instance = provider(activation_scope)
        assert instance.a == 1
        assert instance.b == 2



from prometheus_client import Counter, Gauge, Histogram

CONTROLLER_REQUESTS = Counter(
    "controller_requests_count",
    "Count of controller requests by route name, port and method.",
    ["name", "port", "method"],
)
CONTROLLER_REQUESTS_TIME = Histogram(
    "controller_requests_time",
    "Histogram of controller requests processing time by route name, port and method (in seconds)",
    ["name", "port", "method"],
)
CONTROLLER_REQUESTS_IN_PROGRESS = Gauge(
    "controller_requests_in_progress",
    "Gauge of controller requests by route name, port and method currently being processed",
    ["name", "port", "method"],
)
CONTROLLER_RESPONSES = Counter(
    "controller_responses_count",
    "Count of controller responses by route name, port, method and status and status codes.",
    ["name", "port", "method", "status"],
)
CONTROLLER_EXCEPTIONS = Counter(
    "controller_exceptions_count",
    "Count of exceptions raised in controllers by route name, port, method, path and exception type",
    ["name", "port", "method", "path", "exception"],
)
REMOTE_REQUESTS = Counter(
    "remote_requests_count",
    "Count of remote requests by route name and method.",
    ["name", "method"],
)
REMOTE_REQUESTS_TIME = Histogram(
    "remote_requests_time",
    "Histogram of remote requests processing time by route name and method (in seconds)",
    ["name", "method"],
)
REMOTE_REQUESTS_IN_PROGRESS = Gauge(
    "remote_requests_in_progress",
    "Gauge of remote requests by route name and method currently being processed",
    ["name", "method"],
)
REMOTE_RESPONSES = Counter(
    "remote_responses_count",
    "Count of remote responses by route name, method and status and status codes.",
    ["name", "method", "status"],
)
REMOTE_EXCEPTIONS = Counter(
    "remote_exceptions_count",
    "Count of exceptions raised in remotes by route name, method, path and exception type",
    ["name", "method", "url", "exception"],
)



"""merge 0.6

Revision ID: 4bdd9e1d790d
Revises:
Create Date: 2024-09-06 17:22:11.824809

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "4bdd9e1d790d"
down_revision = ("d4cd37d29d7a", "4ab341a710f4")
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.alter_column(
        "messages", "summary", existing_type=sa.VARCHAR(length=255), type_=sa.Text(), existing_nullable=True
    )


def downgrade() -> None:
    pass



from unittest.mock import AsyncMock
from urllib.parse import urlencode

import pytest
from multidict import MultiDict

from harp.asgi.bridge.requests import HttpRequestAsgiBridge
from harp.http import HttpRequestBridge
from harp.http.utils import HTTP_METHODS

DEFAULT_SERVER_IPADDR = "127.0.0.1"
DEFAULT_SERVER_PORT = 80


unset = object()


class BaseHttpRequestAsgiBridgeTest:
    """
    Base class for testing http requests using an ASGI bridge implementation.
    """

    def create_bridge(self, receive=None, /, **scope) -> HttpRequestBridge:
        return HttpRequestAsgiBridge(
            {
                k: v
                for k, v in {
                    "type": "http",
                    "http_version": "1.1",
                    "asgi": {"spec_version": "2.1", "version": "3.0"},
                    "method": "GET",
                    "scheme": "http",
                    "path": "/",
                    "raw_path": b"/",
                    "query_string": b"",
                    "root_path": "",
                    "headers": [],
                    "client": ["127.0.0.1", 16535],
                    "server": [DEFAULT_SERVER_IPADDR, DEFAULT_SERVER_PORT],
                    "extensions": {},
                    **scope,
                }.items()
                if v is not unset
            },
            receive or AsyncMock(),
        )


class TestHttpRequestAsgiBridgeServer(BaseHttpRequestAsgiBridgeTest):
    """
    Test "server" values from ASGI brigde implementation.
    """

    def test_get_server_default(self):
        bridge = self.create_bridge()
        assert bridge.get_server_ipaddr() == DEFAULT_SERVER_IPADDR
        assert bridge.get_server_port() == DEFAULT_SERVER_PORT

    @pytest.mark.parametrize("server", [None, [], [None], [None, None], unset])
    def test_get_server_invalid_none(self, server):
        bridge = self.create_bridge(server=server)
        assert bridge.get_server_ipaddr() is None
        assert bridge.get_server_port() is None

    @pytest.mark.parametrize("ipaddr", ["127.0.0.1", "172.16.6.1"])
    def test_get_server_override(self, ipaddr):
        bridge = self.create_bridge(server=[ipaddr])
        assert bridge.get_server_ipaddr() == ipaddr
        assert bridge.get_server_port() is None

        bridge = self.create_bridge(server=[ipaddr, 80])
        assert bridge.get_server_ipaddr() == ipaddr
        assert bridge.get_server_port() == 80


class TestHttpRequestAsgiBridgeMethod(BaseHttpRequestAsgiBridgeTest):
    """
    Test HTTP method value from ASGI brigde implementation.
    """

    def test_get_method_default(self):
        bridge = self.create_bridge()
        assert bridge.get_method() == "GET"

    def test_get_method_not_set(self):
        bridge = self.create_bridge(method=unset)
        assert bridge.get_method() is None

    @pytest.mark.parametrize("method", list(HTTP_METHODS.keys()))
    def test_get_method(self, method):
        bridge = self.create_bridge(method=method)
        assert bridge.get_method() == method


class TestHttpRequestAsgiBridgePath(BaseHttpRequestAsgiBridgeTest):
    """
    Test HTTP path (aka ARL, URI, resource ...) value from ASGI brigde implementation.
    """

    def test_get_path_default(self):
        bridge = self.create_bridge()
        assert bridge.get_path() == "/"

    @pytest.mark.parametrize("raw_path", [b"/", b"/foo/bar"])
    def test_get_path_from_raw_path(self, raw_path):
        # raw path has precedence over path
        bridge = self.create_bridge(raw_path=raw_path)
        assert bridge.get_path() == raw_path.decode("utf-8")

        bridge = self.create_bridge(raw_path=raw_path, path=unset)
        assert bridge.get_path() == raw_path.decode("utf-8")

    @pytest.mark.parametrize("path", ["/", "/foo/bar"])
    def test_get_path_from_path(self, path):
        # path is only read if there is no raw path
        bridge = self.create_bridge(path=path, raw_path=unset)
        assert bridge.get_path() == path


class TestHttpRequestAsgiBridgeQuery(BaseHttpRequestAsgiBridgeTest):
    """
    Test HTTP query string value from ASGI brigde implementation.
    """

    def test_get_query_default(self):
        bridge = self.create_bridge()
        assert bridge.get_query() == {}

    def test_get_query_string_basic(self):
        bridge = self.create_bridge(query_string=b"foo=bar&baz=qux")
        assert bridge.get_query() == {"foo": "bar", "baz": "qux"}
        assert urlencode(bridge.get_query()) == "foo=bar&baz=qux"

    def test_get_query_string_multi(self):
        bridge = self.create_bridge(query_string=b"foo=bar&foo=baz&baz=qux")
        assert bridge.get_query() == MultiDict([("foo", "bar"), ("foo", "baz"), ("baz", "qux")])
        assert urlencode(bridge.get_query()) == "foo=bar&foo=baz&baz=qux"

    def test_get_query_string_no_value(self):
        bridge = self.create_bridge(query_string=b"foo")
        assert bridge.get_query() == {"foo": ""}
        assert urlencode(bridge.get_query()) == "foo="


class TestHttpRequestAsgiBridgeHeaders(BaseHttpRequestAsgiBridgeTest):
    """
    Test HTTP headers value from ASGI brigde implementation.
    """

    def test_get_headers_default(self):
        bridge = self.create_bridge()
        assert bridge.get_headers() == {}

    def test_get_headers(self):
        bridge = self.create_bridge(headers=[(b"content-type", b"application/json"), (b"content-length", b"42")])
        assert bridge.get_headers() == {
            "content-type": "application/json",
            "content-length": "42",
        }


class TestHttpRequestAsgiBridgeBody(BaseHttpRequestAsgiBridgeTest):
    """
    Test HTTP body value from ASGI brigde implementation. This is one of the most comlpex parts of the bridge, as the
    ASGI protocol allows for streaming the body in chunks, thus we need to access this asynchronously.
    """

    def create_receive_stub(self, *messages):
        messages = list(messages)

        async def receive():
            return messages.pop(0)

        return receive, messages

    async def test_body_basic(self):
        receive, messages = self.create_receive_stub({"type": "http.request", "body": b'{"foo": "bar"}'})
        bridge = self.create_bridge(receive)

        chunks = []
        assert len(messages)
        async for chunk in bridge.get_stream():
            chunks.append(chunk)

        assert b"".join(chunks) == b'{"foo": "bar"}'
        assert len(chunks) == 1
        assert not len(messages)

    async def test_body_multi_chunk(self):
        receive, messages = self.create_receive_stub(
            {"type": "http.request", "body": b'{"foo":', "more_body": True},
            {"type": "http.request", "body": b' "bar"}'},
        )
        bridge = self.create_bridge(receive)

        chunks = []
        assert len(messages)
        async for chunk in bridge.get_stream():
            chunks.append(chunk)

        assert b"".join(chunks) == b'{"foo": "bar"}'
        assert len(chunks) == 2
        assert not len(messages)



import asyncio
from typing import cast

import click
from whistle import IAsyncEventDispatcher

from harp.commandline.options.server import config_command
from harp.config import ConfigurationBuilder
from harp.event_dispatcher import LoggingAsyncEventDispatcher
from harp.http import HttpRequest
from harp.utils.urls import normalize_url
from harp_apps.http_client.events import EVENT_FILTER_HTTP_CLIENT_REQUEST, EVENT_FILTER_HTTP_CLIENT_RESPONSE
from harp_apps.proxy.controllers import HttpProxyController
from harp_apps.proxy.events import EVENT_FILTER_PROXY_REQUEST, EVENT_FILTER_PROXY_RESPONSE
from harp_apps.proxy.settings import Remote

from .utils.dump import (
    on_proxy_request_dump,
    on_proxy_response_dump,
    on_remote_request_dump,
    on_remote_response_dump,
    on_remote_response_show_cache_control,
)
from .utils.loaders import load_ruleset_from_files
from .utils.subscribers import DebugRulesSubscriber


@config_command("run")
@click.argument("endpoint")
@click.argument("method")
@click.argument("path")
def run_command(files, examples, options, endpoint, method, path):
    try:
        endpoint_name, endpoint_target = endpoint.split("=", 1)
    except ValueError:
        raise click.BadParameter("Endpoint must be in the format name=url")
    endpoint_target = normalize_url(endpoint_target)

    ruleset = load_ruleset_from_files(files, examples, options)

    dispatcher: IAsyncEventDispatcher = cast(IAsyncEventDispatcher, LoggingAsyncEventDispatcher())

    system = asyncio.run(
        ConfigurationBuilder({"applications": ["http_client", "rules"]}, use_default_applications=False).abuild_system()
    )

    http_client = system.provider.get("http_client")

    dispatcher.add_listener(EVENT_FILTER_PROXY_REQUEST, on_proxy_request_dump, priority=-100)
    dispatcher.add_listener(EVENT_FILTER_HTTP_CLIENT_REQUEST, on_remote_request_dump, priority=-100)
    dispatcher.add_listener(EVENT_FILTER_HTTP_CLIENT_RESPONSE, on_remote_response_dump, priority=-100)
    dispatcher.add_listener(
        EVENT_FILTER_HTTP_CLIENT_RESPONSE,
        on_remote_response_show_cache_control,
        priority=-100,
    )
    dispatcher.add_listener(
        EVENT_FILTER_HTTP_CLIENT_RESPONSE,
        on_remote_response_show_cache_control,
        priority=100,
    )
    dispatcher.add_listener(EVENT_FILTER_PROXY_RESPONSE, on_proxy_response_dump, priority=-100)

    # Create a proxy controller to mimic the real behavious of the proxy.
    controller = HttpProxyController(
        Remote.from_settings_dict({"endpoints": [{"url": endpoint_target}]}),
        name=endpoint_name,
        dispatcher=dispatcher,
        http_client=http_client,
    )

    rules_subscriber = DebugRulesSubscriber(ruleset)
    rules_subscriber.subscribe(dispatcher)

    request = HttpRequest(method=method, path=path)
    asyncio.run(controller(request))



from datetime import UTC, datetime, timedelta

import pytest

from harp_apps.janitor.settings import OLD_AFTER
from harp_apps.janitor.worker import JanitorWorker
from harp_apps.storage.services.sql import SqlStorage
from harp_apps.storage.types import IBlobStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class TestJanitorWorker(StorageTestFixtureMixin):
    async def test_delete_old_transactions(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        worker = JanitorWorker(sql_storage, blob_storage)

        await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(hours=1))
        await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(minutes=1))
        await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER + timedelta(minutes=1))

        async with sql_storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 3

        await worker.delete_old_transactions()

        async with sql_storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 1

    async def test_delete_orphan_blobs(self, sql_storage: SqlStorage, blob_storage: IBlobStorage):
        if blob_storage.type == "redis":
            # explicit skip because it should be implemented later
            return pytest.skip("Redis implementation does not support cleaning up orphan blobs yet.")

        worker = JanitorWorker(sql_storage, blob_storage)

        b1 = await self.create_blob(blob_storage, "foo")
        b2 = await self.create_blob(blob_storage, "bar")
        await self.create_blob(blob_storage, "baz")

        async with sql_storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 3
            assert metrics["storage.blobs.orphans"] == 3

        t = await self.create_transaction(sql_storage)
        await self.create_message(
            sql_storage,
            transaction_id=t.id,
            kind="misc",
            summary="foo",
            headers=b1.id,
            body=b2.id,
        )

        async with sql_storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 3
            assert metrics["storage.blobs.orphans"] == 1

        await worker.delete_orphan_blobs()

        async with sql_storage.session_factory() as session:
            metrics = await worker.compute_metrics(session)
            assert metrics["storage.blobs"] == 2
            assert metrics["storage.blobs.orphans"] == 0

    async def test_delete_old_transactions_but_keep_flagged_ones(
        self, sql_storage: SqlStorage, blob_storage: IBlobStorage
    ):
        worker = JanitorWorker(sql_storage, blob_storage)

        t1 = await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(hours=1))
        await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER - timedelta(minutes=1))
        await self.create_transaction(sql_storage, started_at=datetime.now(UTC) - OLD_AFTER + timedelta(minutes=1))

        user = await sql_storage.users.find_one_by_username("anonymous")
        await sql_storage.flags.create({"type": 1, "user_id": user.id, "transaction_id": t1.id})

        async with sql_storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 3

        await worker.delete_old_transactions()

        async with sql_storage.session_factory() as session:
            assert (await worker.compute_metrics(session))["storage.transactions"] == 2



import os.path
from glob import glob
from itertools import chain

import pytest

import harp
from harp.config.asdict import asdict
from harp.config.examples import get_available_examples, get_example_filename
from harp.utils.config import yaml


def _get_available_documentation_examples_filenames():
    files = chain(
        *(
            glob(f"docs/**/examples/*.{format}", root_dir=harp.ROOT_DIR, recursive=True)
            for format in ("yml", "yaml", "toml")
        )
    )

    return list(sorted(files))


def test_get_available_examples(snapshot):
    assert get_available_examples() == snapshot


def test_documentation_examples_list(snapshot):
    assert [
        x.removeprefix(
            harp.ROOT_DIR + "/",
        )
        for x in _get_available_documentation_examples_filenames()
    ] == snapshot


@pytest.mark.parametrize("example", get_available_examples())
def test_load_example(example, snapshot):
    from harp.config import ConfigurationBuilder

    builder = ConfigurationBuilder()
    builder.add_file(get_example_filename(example))
    settings = builder()
    assert yaml.dump(asdict(settings)) == snapshot


@pytest.mark.parametrize("configfile", _get_available_documentation_examples_filenames())
def test_load_documentation_example(configfile, snapshot):
    from harp.config import ConfigurationBuilder

    builder = ConfigurationBuilder()
    builder.add_file(os.path.join(harp.ROOT_DIR, configfile))
    settings = builder()
    applications = settings.pop("applications", [])
    assert len(set(settings.keys()).difference(set(builder.applications.keys()))) == 0
    assert len(applications) >= len(settings)
    assert yaml.dump(asdict(settings)) == snapshot



from harp.commandline.examples import list_command
from harp.utils.testing.cli import CliRunner


def test_list_examples_command(snapshot):
    runner = CliRunner()

    result = runner.invoke(list_command, [])
    assert result.exit_code == 0
    assert result.output.strip() == snapshot

    result = runner.invoke(list_command, ["--raw"])
    assert result.exit_code == 0
    assert result.output.strip() == snapshot
